from discord_self._vendor.discord.commands import (
    BaseCommand,
    MessageCommand,
    Option,
    OptionChoice,
    SlashCommand,
    SubCommand,
    UserCommand,
)

__all__ = [
    "BaseCommand",
    "MessageCommand",
    "Option",
    "OptionChoice",
    "SlashCommand",
    "SubCommand",
    "UserCommand",
]



from discord_self._vendor.discord.webhook.async_ import (
    PartialWebhookChannel,
    PartialWebhookGuild,
    Webhook,
    WebhookMessage,
)

__all__ = ["PartialWebhookChannel", "PartialWebhookGuild", "Webhook", "WebhookMessage"]



from discord_self._vendor.discord.types.embed import (
    Embed,
    EmbedAuthor,
    EmbedField,
    EmbedFooter,
    EmbedImage,
    EmbedProvider,
    EmbedThumbnail,
    EmbedType,
    EmbedVideo,
)

__all__ = [
    "Embed",
    "EmbedAuthor",
    "EmbedField",
    "EmbedFooter",
    "EmbedImage",
    "EmbedProvider",
    "EmbedThumbnail",
    "EmbedType",
    "EmbedVideo",
]



from discord_self._vendor.discord.types.application import (
    ApplicationInstallParams,
    RoleConnection,
)
from discord_self._vendor.discord.types.member import PrivateMember as ProfileMember
from discord_self._vendor.discord.types.profile import (
    MutualGuild,
    Profile,
    ProfileApplication,
    ProfileMetadata,
    ProfileUser,
)
from discord_self._vendor.discord.types.snowflake import Snowflake
from discord_self._vendor.discord.types.user import (
    APIUser,
    PartialConnection,
    PremiumType,
)

__all__ = [
    "APIUser",
    "ApplicationInstallParams",
    "MutualGuild",
    "PartialConnection",
    "PremiumType",
    "Profile",
    "ProfileApplication",
    "ProfileMember",
    "ProfileMetadata",
    "ProfileUser",
    "RoleConnection",
    "Snowflake",
]



from discord_self._vendor.discord.message import (
    Attachment,
    DeletedReferencedMessage,
    Message,
    MessageReference,
    PartialMessage,
    RoleSubscriptionInfo,
)

__all__ = [
    "Attachment",
    "DeletedReferencedMessage",
    "Message",
    "MessageReference",
    "PartialMessage",
    "RoleSubscriptionInfo",
]



from discord_self._vendor.discord.types.automod import (
    AutoModerationAction,
    AutoModerationActionExecution,
    AutoModerationActionTriggerType,
    AutoModerationRule,
    AutoModerationRuleEventType,
    AutoModerationRuleTriggerType,
    AutoModerationTriggerMetadata,
    AutoModerationTriggerPresets,
    Empty,
)
from discord_self._vendor.discord.types.snowflake import Snowflake

__all__ = [
    "AutoModerationAction",
    "AutoModerationActionExecution",
    "AutoModerationActionTriggerType",
    "AutoModerationRule",
    "AutoModerationRuleEventType",
    "AutoModerationRuleTriggerType",
    "AutoModerationTriggerMetadata",
    "AutoModerationTriggerPresets",
    "Empty",
    "Snowflake",
]



from discord_self._vendor.discord.types.activity import PartialPresenceUpdate
from discord_self._vendor.discord.types.channel import GuildChannel, StageInstance
from discord_self._vendor.discord.types.emoji import Emoji
from discord_self._vendor.discord.types.guild import (
    ApplicationCommandCounts,
    Ban,
    BaseGuild,
    ChannelPositionUpdate,
    DefaultMessageNotificationLevel,
    ExplicitContentFilterLevel,
    Guild,
    GuildMFALevel,
    GuildPreview,
    GuildPrune,
    GuildWithCounts,
    InviteGuild,
    MFALevel,
    NSFWLevel,
    PartialGuild,
    PremiumTier,
    RolePositionUpdate,
    SupplementalGuild,
    UnavailableGuild,
    UserGuild,
    VerificationLevel,
)
from discord_self._vendor.discord.types.member import MemberWithUser
from discord_self._vendor.discord.types.role import Role
from discord_self._vendor.discord.types.scheduled_event import GuildScheduledEvent
from discord_self._vendor.discord.types.snowflake import Snowflake
from discord_self._vendor.discord.types.sticker import GuildSticker
from discord_self._vendor.discord.types.threads import Thread
from discord_self._vendor.discord.types.user import User
from discord_self._vendor.discord.types.voice import GuildVoiceState
from discord_self._vendor.discord.types.welcome_screen import WelcomeScreen

__all__ = [
    "ApplicationCommandCounts",
    "Ban",
    "BaseGuild",
    "ChannelPositionUpdate",
    "DefaultMessageNotificationLevel",
    "Emoji",
    "ExplicitContentFilterLevel",
    "Guild",
    "GuildChannel",
    "GuildMFALevel",
    "GuildPreview",
    "GuildPrune",
    "GuildScheduledEvent",
    "GuildSticker",
    "GuildVoiceState",
    "GuildWithCounts",
    "InviteGuild",
    "MFALevel",
    "MemberWithUser",
    "NSFWLevel",
    "PartialGuild",
    "PartialPresenceUpdate",
    "PremiumTier",
    "Role",
    "RolePositionUpdate",
    "Snowflake",
    "StageInstance",
    "SupplementalGuild",
    "Thread",
    "UnavailableGuild",
    "User",
    "UserGuild",
    "VerificationLevel",
    "WelcomeScreen",
]



from discord_self._vendor.discord.components import (
    ActionRow,
    Button,
    Component,
    SelectMenu,
    SelectOption,
    TextInput,
)

__all__ = [
    "ActionRow",
    "Button",
    "Component",
    "SelectMenu",
    "SelectOption",
    "TextInput",
]



from discord_self._vendor.discord import utils
from discord_self._vendor.discord.channel import (
    DMChannel,
    ForumChannel,
    GroupChannel,
    PartialMessageable,
    TextChannel,
    VoiceChannel,
)
from discord_self._vendor.discord.embeds import Embed
from discord_self._vendor.discord.enums import (
    AuditLogAction,
    ChannelType,
    InteractionType,
    InviteType,
    RelationshipAction,
)
from discord_self._vendor.discord.errors import (
    CaptchaRequired,
    DiscordServerError,
    Forbidden,
    GatewayNotFound,
    HTTPException,
    LoginFailure,
    NotFound,
    RateLimited,
)
from discord_self._vendor.discord.file import File
from discord_self._vendor.discord.flags import MessageFlags
from discord_self._vendor.discord.handlers import CaptchaHandler
from discord_self._vendor.discord.http import (
    INTERNAL_API_VERSION,
    HTTPClient,
    MultipartParameters,
    Ratelimit,
    Route,
    handle_message_parameters,
    json_or_text,
)
from discord_self._vendor.discord.mentions import AllowedMentions
from discord_self._vendor.discord.message import Attachment, Message
from discord_self._vendor.discord.threads import Thread
from discord_self._vendor.discord.tracking import ContextProperties
from discord_self._vendor.discord.types import (
    application,
    audit_log,
    automod,
    billing,
    channel,
    emoji,
    entitlements,
    guild,
    integration,
    invite,
    library,
    member,
    message,
    payments,
    profile,
    promotions,
    role,
    scheduled_event,
    sticker,
    store,
    subscriptions,
    team,
    template,
    threads,
    user,
    webhook,
    welcome_screen,
    widget,
)
from discord_self._vendor.discord.types.snowflake import Snowflake, SnowflakeList
from discord_self._vendor.discord.utils import MISSING

__all__ = [
    "AllowedMentions",
    "Attachment",
    "AuditLogAction",
    "CaptchaHandler",
    "CaptchaRequired",
    "ChannelType",
    "ContextProperties",
    "DMChannel",
    "DiscordServerError",
    "Embed",
    "File",
    "Forbidden",
    "ForumChannel",
    "GatewayNotFound",
    "GroupChannel",
    "HTTPClient",
    "HTTPException",
    "INTERNAL_API_VERSION",
    "InteractionType",
    "InviteType",
    "LoginFailure",
    "MISSING",
    "Message",
    "MessageFlags",
    "MultipartParameters",
    "NotFound",
    "PartialMessageable",
    "RateLimited",
    "Ratelimit",
    "RelationshipAction",
    "Route",
    "Snowflake",
    "SnowflakeList",
    "TextChannel",
    "Thread",
    "VoiceChannel",
    "application",
    "audit_log",
    "automod",
    "billing",
    "channel",
    "emoji",
    "entitlements",
    "guild",
    "handle_message_parameters",
    "integration",
    "invite",
    "json_or_text",
    "library",
    "member",
    "message",
    "payments",
    "profile",
    "promotions",
    "role",
    "scheduled_event",
    "sticker",
    "store",
    "subscriptions",
    "team",
    "template",
    "threads",
    "user",
    "utils",
    "webhook",
    "welcome_screen",
    "widget",
]



from discord_self._vendor.discord.types.billing import (
    BillingAddress,
    BillingAddressToken,
    PartialPaymentSource,
    PaymentSource,
    PremiumUsage,
    PremiumUsageValue,
)

__all__ = [
    "BillingAddress",
    "BillingAddressToken",
    "PartialPaymentSource",
    "PaymentSource",
    "PremiumUsage",
    "PremiumUsageValue",
]



from discord_self._vendor.discord.settings import (
    AudioContext,
    ChannelSettings,
    EmailSettings,
    GuildFolder,
    GuildProgress,
    GuildSettings,
    LegacyUserSettings,
    MuteConfig,
    TrackingSettings,
    UserSettings,
)

__all__ = [
    "AudioContext",
    "ChannelSettings",
    "EmailSettings",
    "GuildFolder",
    "GuildProgress",
    "GuildSettings",
    "LegacyUserSettings",
    "MuteConfig",
    "TrackingSettings",
    "UserSettings",
]



from discord_self._vendor.discord.flags import (
    ApplicationDiscoveryFlags,
    ApplicationFlags,
    AutoModPresets,
    Capabilities,
    ChannelFlags,
    FriendDiscoveryFlags,
    FriendSourceFlags,
    GiftFlags,
    HubProgressFlags,
    LibraryApplicationFlags,
    MemberCacheFlags,
    MemberFlags,
    MessageFlags,
    OnboardingProgressFlags,
    PaymentFlags,
    PaymentSourceFlags,
    PremiumUsageFlags,
    PrivateUserFlags,
    PromotionFlags,
    PublicUserFlags,
    PurchasedFlags,
    SKUFlags,
    SystemChannelFlags,
)

__all__ = [
    "ApplicationDiscoveryFlags",
    "ApplicationFlags",
    "AutoModPresets",
    "Capabilities",
    "ChannelFlags",
    "FriendDiscoveryFlags",
    "FriendSourceFlags",
    "GiftFlags",
    "HubProgressFlags",
    "LibraryApplicationFlags",
    "MemberCacheFlags",
    "MemberFlags",
    "MessageFlags",
    "OnboardingProgressFlags",
    "PaymentFlags",
    "PaymentSourceFlags",
    "PremiumUsageFlags",
    "PrivateUserFlags",
    "PromotionFlags",
    "PublicUserFlags",
    "PurchasedFlags",
    "SKUFlags",
    "SystemChannelFlags",
]



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional, Union

from .object import Object
from .partial_emoji import PartialEmoji
from .utils import MISSING, _get_as_snowflake

if TYPE_CHECKING:
    from .abc import Snowflake
    from .emoji import Emoji
    from .guild import Guild
    from .invite import PartialInviteGuild
    from .state import ConnectionState
    from .types.welcome_screen import (
        WelcomeScreen as WelcomeScreenPayload,
    )
    from .types.welcome_screen import (
        WelcomeScreenChannel as WelcomeScreenChannelPayload,
    )

__all__ = (
    "WelcomeChannel",
    "WelcomeScreen",
)


class WelcomeChannel:
    """Represents a channel shown on a :class:`WelcomeScreen`.

    .. versionadded:: 2.0

    Attributes
    -----------
    channel: :class:`abc.Snowflake`
        The channel that is being shown.
    description: :class:`str`
        The description of the channel.
    emoji: Optional[Union[:class:`PartialEmoji`, :class:`Emoji`]
        The emoji shown under the description.
    """

    def __init__(
        self,
        *,
        channel: Snowflake,
        description: str,
        emoji: Optional[Union[PartialEmoji, Emoji, str]] = None,
    ) -> None:
        self.channel = channel
        self.description = description

        if isinstance(emoji, str):
            emoji = PartialEmoji(name=emoji)
        self.emoji = emoji

    def __repr__(self) -> str:
        return f"<WelcomeChannel channel={self.channel!r} description={self.description} emoji={self.emoji!r}>"

    @classmethod
    def _from_dict(
        cls, *, data: WelcomeScreenChannelPayload, state: ConnectionState
    ) -> WelcomeChannel:
        channel_id = int(data["channel_id"])
        channel = state.get_channel(channel_id) or Object(id=channel_id)

        emoji = None
        if (emoji_id := _get_as_snowflake(data, "emoji_id")) is not None:
            emoji = state.get_emoji(emoji_id)
        elif (emoji_name := data.get("emoji_name")) is not None:
            emoji = PartialEmoji(name=emoji_name)

        return cls(
            channel=channel, description=data.get("description", ""), emoji=emoji
        )

    def _to_dict(self) -> WelcomeScreenChannelPayload:
        data: WelcomeScreenChannelPayload = {
            "channel_id": self.channel.id,
            "description": self.description,
            "emoji_id": None,
            "emoji_name": None,
        }
        if (emoji := self.emoji) is not None:
            data["emoji_id"] = emoji.id
            data["emoji_name"] = emoji.name

        return data


class WelcomeScreen:
    """Represents a :class:`Guild`'s welcome screen.

    .. versionadded:: 2.0

    .. container:: operations

        .. describe:: bool(b)

            Returns whether the welcome screen is enabled.

    Attributes
    -----------
    guild: Union[:class:`Guild`, :class:`PartialInviteGuild`]
        The guild the welcome screen is for.
    description: :class:`str`
        The text shown on the welcome screen.
    welcome_channels: List[:class:`WelcomeChannel`]
        The channels shown on the welcome screen.
    """

    def __init__(
        self, *, data: WelcomeScreenPayload, guild: Union[Guild, PartialInviteGuild]
    ) -> None:
        self.guild = guild
        self._update(data)

    def _update(self, data: WelcomeScreenPayload) -> None:
        state = self.guild._state
        channels = data.get("welcome_channels", [])

        self.welcome_channels: List[WelcomeChannel] = [
            WelcomeChannel._from_dict(data=channel, state=state) for channel in channels
        ]
        self.description: str = data.get("description", "")

    def __repr__(self) -> str:
        return f"<WelcomeScreen enabled={self.enabled} description={self.description} welcome_channels={self.welcome_channels!r}>"

    def __bool__(self) -> bool:
        return self.enabled

    @property
    def enabled(self) -> bool:
        """:class:`bool`: Whether the welcome screen is displayed."""
        return "WELCOME_SCREEN_ENABLED" in self.guild.features

    async def edit(
        self,
        *,
        description: str = MISSING,
        welcome_channels: List[WelcomeChannel] = MISSING,
        enabled: bool = MISSING,
        reason: Optional[str] = None,
    ):
        """|coro|

        Edit the welcome screen.

        Welcome channels can only accept custom emojis if :attr:`Guild.premium_tier` is level 2 or above.

        You must have :attr:`~Permissions.manage_guild` in the guild to do this.

        All parameters are optional.

        Usage: ::

            rules_channel = guild.get_channel(12345678)
            announcements_channel = guild.get_channel(87654321)

            custom_emoji = utils.get(guild.emojis, name='loudspeaker')

            await welcome_screen.edit(
                description='This is a very cool community server!',
                welcome_channels=[
                    WelcomeChannel(channel=rules_channel, description='Read the rules!', emoji='👨‍🏫'),
                    WelcomeChannel(channel=announcements_channel, description='Watch out for announcements!', emoji=custom_emoji),
                ]
            )

        Parameters
        ------------
        enabled: :class:`bool`
            Whether the welcome screen will be shown.
        description: :class:`str`
            The welcome screen's description.
        welcome_channels: Optional[List[:class:`WelcomeChannel`]]
            The welcome channels (in order).
        reason: Optional[:class:`str`]
            The reason for editing the welcome screen. Shows up on the audit log.

        Raises
        -------
        HTTPException
            Editing the welcome screen failed failed.
        Forbidden
            You don't have permissions to edit the welcome screen.
        """
        payload = {}

        if enabled is not MISSING:
            payload["enabled"] = enabled
        if description is not MISSING:
            payload["description"] = description
        if welcome_channels is not MISSING:
            channels = (
                [channel._to_dict() for channel in welcome_channels]
                if welcome_channels
                else []
            )
            payload["welcome_channels"] = channels

        if payload:
            guild = self.guild
            data = await guild._state.http.edit_welcome_screen(
                guild.id, payload, reason=reason
            )
            self._update(data)



from discord_self._vendor.discord.types.promotions import (
    ClaimedPromotion,
    PricingPromotion,
    Promotion,
    PromotionalPrice,
    TrialOffer,
    WrappedPricingPromotion,
)
from discord_self._vendor.discord.types.snowflake import Snowflake
from discord_self._vendor.discord.types.subscriptions import SubscriptionTrial

__all__ = [
    "ClaimedPromotion",
    "PricingPromotion",
    "Promotion",
    "PromotionalPrice",
    "Snowflake",
    "SubscriptionTrial",
    "TrialOffer",
    "WrappedPricingPromotion",
]



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional, Union

from .components import _component_factory
from .enums import InteractionType
from .interactions import _wrapped_interaction
from .mixins import Hashable
from .utils import _generate_nonce

if TYPE_CHECKING:
    from .application import IntegrationApplication
    from .components import ActionRow
    from .interactions import Interaction

# fmt: off
__all__ = (
    'Modal',
)
# fmt: on


class Modal(Hashable):
    """Represents a modal from the Discord Bot UI Kit.

    .. versionadded:: 2.0

    .. container:: operations

        .. describe:: x == y

            Checks if two modals are equal.

        .. describe:: x != y

            Checks if two modals are not equal.

        .. describe:: hash(x)

            Return the modal's hash.

        .. describe:: str(x)

            Returns the modal's title.

    Attributes
    -----------
    id: :class:`int`
        The modal's ID. This is the same as the interaction ID.
    nonce: Optional[Union[:class:`int`, :class:`str`]]
        The modal's nonce. May not be present.
    title: :class:`str`
        The modal's title.
    custom_id: :class:`str`
        The ID of the modal that gets received during an interaction.
    components: List[:class:`Component`]
        A list of components in the modal.
    application: :class:`IntegrationApplication`
        The application that sent the modal.
    """

    __slots__ = (
        "_state",
        "interaction",
        "id",
        "nonce",
        "title",
        "custom_id",
        "components",
        "application",
    )

    def __init__(self, *, data: dict, interaction: Interaction):
        self._state = interaction._state
        self.interaction = interaction
        self.id = int(data["id"])
        self.nonce: Optional[Union[int, str]] = data.get("nonce")
        self.title: str = data.get("title", "")
        self.custom_id: str = data.get("custom_id", "")
        self.components: List[ActionRow] = [
            _component_factory(d) for d in data.get("components", [])
        ]  # type: ignore # Will always be rows here
        self.application: IntegrationApplication = (
            interaction._state.create_integration_application(data["application"])
        )

    def __str__(self) -> str:
        return self.title

    def to_dict(self) -> dict:
        return {
            "id": str(self.id),
            "custom_id": self.custom_id,
            "components": [c.to_dict() for c in self.components],
        }

    async def submit(self):
        """|coro|

        Submits the modal.

        All required components must be already answered.

        Raises
        -------
        InvalidData
            Didn't receive a response from Discord
            (doesn't mean the interaction failed).
        NotFound
            The originating message was not found.
        HTTPException
            Choosing the options failed.

        Returns
        --------
        :class:`Interaction`
            The interaction that was created.
        """
        interaction = self.interaction
        return await _wrapped_interaction(
            self._state,
            _generate_nonce(),
            InteractionType.modal_submit,
            None,
            interaction.channel,
            self.to_dict(),
            application_id=self.application.id,
        )



from discord_self._vendor.discord import (
    __author__,
    __copyright__,
    __license__,
    __path__,
    __title__,
    __version__,
    abc,
    opus,
    utils,
    version_info,
)
from discord_self._vendor.discord.activity import (
    Activity,
    BaseActivity,
    CustomActivity,
    Game,
    Session,
    Spotify,
    Streaming,
)
from discord_self._vendor.discord.affinity import GuildAffinity, UserAffinity
from discord_self._vendor.discord.application import (
    EULA,
    Achievement,
    Application,
    ApplicationActivityStatistics,
    ApplicationAsset,
    ApplicationBot,
    ApplicationBranch,
    ApplicationBuild,
    ApplicationExecutable,
    ApplicationInstallParams,
    ApplicationTester,
    Company,
    EmbeddedActivityConfig,
    IntegrationApplication,
    Manifest,
    ManifestLabel,
    PartialApplication,
    ThirdPartySKU,
)
from discord_self._vendor.discord.asset import Asset
from discord_self._vendor.discord.audit_logs import (
    AuditLogChanges,
    AuditLogDiff,
    AuditLogEntry,
)
from discord_self._vendor.discord.automod import (
    AutoModAction,
    AutoModRule,
    AutoModRuleAction,
    AutoModTrigger,
)
from discord_self._vendor.discord.billing import (
    BillingAddress,
    PaymentSource,
    PremiumUsage,
)
from discord_self._vendor.discord.calls import CallMessage, GroupCall, PrivateCall
from discord_self._vendor.discord.channel import (
    CategoryChannel,
    DMChannel,
    ForumChannel,
    ForumTag,
    GroupChannel,
    PartialMessageable,
    StageChannel,
    TextChannel,
    VoiceChannel,
)
from discord_self._vendor.discord.client import Client
from discord_self._vendor.discord.colour import Color, Colour
from discord_self._vendor.discord.commands import (
    BaseCommand,
    MessageCommand,
    Option,
    OptionChoice,
    SlashCommand,
    SubCommand,
    UserCommand,
)
from discord_self._vendor.discord.components import (
    ActionRow,
    Button,
    Component,
    SelectMenu,
    SelectOption,
    TextInput,
)
from discord_self._vendor.discord.connections import Connection, PartialConnection
from discord_self._vendor.discord.embeds import Embed
from discord_self._vendor.discord.emoji import Emoji
from discord_self._vendor.discord.entitlements import Entitlement, Gift, GiftBatch
from discord_self._vendor.discord.enums import (
    ActivityType,
    AppCommandOptionType,
    AppCommandType,
    ApplicationAssetType,
    ApplicationDiscoverabilityState,
    ApplicationMembershipState,
    ApplicationType,
    ApplicationVerificationState,
    AuditLogAction,
    AuditLogActionCategory,
    AutoModRuleActionType,
    AutoModRuleEventType,
    AutoModRuleTriggerType,
    ButtonStyle,
    ChannelType,
    ClientType,
    ComponentType,
    ConnectionType,
    ContentFilter,
    ContentRatingAgency,
    DefaultAvatar,
    Distributor,
    EmbeddedActivityOrientation,
    EmbeddedActivityPlatform,
    EmojiPickerSection,
    EntitlementType,
    EntityType,
    EventStatus,
    ExpireBehavior,
    ExpireBehaviour,
    ForumLayoutType,
    ForumOrderType,
    GiftStyle,
    HighlightLevel,
    HypeSquadHouse,
    InboxTab,
    InteractionType,
    InviteTarget,
    InviteType,
    Locale,
    MessageType,
    MFALevel,
    NotificationLevel,
    NSFWLevel,
    OperatingSystem,
    PaymentGateway,
    PaymentSourceType,
    PaymentStatus,
    PayoutAccountStatus,
    PayoutReportType,
    PayoutStatus,
    PremiumType,
    PrivacyLevel,
    RelationshipType,
    ReportType,
    RequiredActionType,
    RPCApplicationState,
    ScheduledEventEntityType,
    ScheduledEventStatus,
    SKUAccessLevel,
    SKUFeature,
    SKUGenre,
    SKUType,
    SpeakingState,
    SpoilerRenderOptions,
    Status,
    StickerAnimationOptions,
    StickerFormatType,
    StickerPickerSection,
    StickerType,
    StoreApplicationState,
    SubscriptionDiscountType,
    SubscriptionInterval,
    SubscriptionInvoiceStatus,
    SubscriptionPlanPurchaseType,
    SubscriptionStatus,
    SubscriptionType,
    TextStyle,
    Theme,
    UserContentFilter,
    UserFlags,
    VerificationLevel,
    VideoQualityMode,
    WebhookType,
)
from discord_self._vendor.discord.errors import (
    AuthFailure,
    CaptchaRequired,
    ClientException,
    ConnectionClosed,
    DiscordException,
    DiscordServerError,
    Forbidden,
    GatewayNotFound,
    HTTPException,
    InvalidData,
    LoginFailure,
    NotFound,
    RateLimited,
)
from discord_self._vendor.discord.file import File
from discord_self._vendor.discord.flags import (
    ApplicationDiscoveryFlags,
    ApplicationFlags,
    AutoModPresets,
    Capabilities,
    ChannelFlags,
    FriendDiscoveryFlags,
    FriendSourceFlags,
    GiftFlags,
    HubProgressFlags,
    LibraryApplicationFlags,
    MemberCacheFlags,
    MemberFlags,
    MessageFlags,
    OnboardingProgressFlags,
    PaymentFlags,
    PaymentSourceFlags,
    PremiumUsageFlags,
    PrivateUserFlags,
    PromotionFlags,
    PublicUserFlags,
    PurchasedFlags,
    SKUFlags,
    SystemChannelFlags,
)
from discord_self._vendor.discord.guild import (
    ApplicationCommandCounts,
    BanEntry,
    Guild,
    UserGuild,
)
from discord_self._vendor.discord.guild_premium import (
    PremiumGuildSubscription,
    PremiumGuildSubscriptionCooldown,
    PremiumGuildSubscriptionSlot,
)
from discord_self._vendor.discord.handlers import CaptchaHandler
from discord_self._vendor.discord.integrations import (
    BotIntegration,
    Integration,
    IntegrationAccount,
    StreamIntegration,
)
from discord_self._vendor.discord.interactions import Interaction
from discord_self._vendor.discord.invite import (
    Invite,
    PartialInviteChannel,
    PartialInviteGuild,
)
from discord_self._vendor.discord.library import LibraryApplication, LibrarySKU
from discord_self._vendor.discord.member import Member, VoiceState
from discord_self._vendor.discord.mentions import AllowedMentions
from discord_self._vendor.discord.message import (
    Attachment,
    DeletedReferencedMessage,
    Message,
    MessageReference,
    PartialMessage,
    RoleSubscriptionInfo,
)
from discord_self._vendor.discord.metadata import Metadata
from discord_self._vendor.discord.modal import Modal
from discord_self._vendor.discord.object import Object
from discord_self._vendor.discord.partial_emoji import PartialEmoji
from discord_self._vendor.discord.payments import EntitlementPayment, Payment
from discord_self._vendor.discord.permissions import PermissionOverwrite, Permissions
from discord_self._vendor.discord.player import (
    AudioSource,
    FFmpegAudio,
    FFmpegOpusAudio,
    FFmpegPCMAudio,
    PCMAudio,
    PCMVolumeTransformer,
)
from discord_self._vendor.discord.profile import (
    ApplicationProfile,
    MemberProfile,
    MutualGuild,
    UserProfile,
)
from discord_self._vendor.discord.promotions import (
    PricingPromotion,
    Promotion,
    TrialOffer,
)
from discord_self._vendor.discord.raw_models import (
    RawBulkMessageDeleteEvent,
    RawIntegrationDeleteEvent,
    RawMessageDeleteEvent,
    RawMessageUpdateEvent,
    RawReactionActionEvent,
    RawReactionClearEmojiEvent,
    RawReactionClearEvent,
    RawThreadDeleteEvent,
    RawThreadMembersUpdate,
)
from discord_self._vendor.discord.reaction import Reaction
from discord_self._vendor.discord.relationship import Relationship
from discord_self._vendor.discord.role import Role, RoleTags
from discord_self._vendor.discord.scheduled_event import ScheduledEvent
from discord_self._vendor.discord.settings import (
    AudioContext,
    ChannelSettings,
    EmailSettings,
    GuildFolder,
    GuildProgress,
    GuildSettings,
    LegacyUserSettings,
    MuteConfig,
    TrackingSettings,
    UserSettings,
)
from discord_self._vendor.discord.stage_instance import StageInstance
from discord_self._vendor.discord.sticker import (
    GuildSticker,
    StandardSticker,
    Sticker,
    StickerItem,
    StickerPack,
)
from discord_self._vendor.discord.store import (
    SKU,
    ContentRating,
    SKUPrice,
    StoreAsset,
    StoreListing,
    StoreNote,
    SubscriptionPlan,
    SubscriptionPlanPrices,
    SystemRequirements,
)
from discord_self._vendor.discord.subscriptions import (
    Subscription,
    SubscriptionDiscount,
    SubscriptionInvoice,
    SubscriptionInvoiceItem,
    SubscriptionItem,
    SubscriptionRenewalMutations,
    SubscriptionTrial,
)
from discord_self._vendor.discord.team import Team, TeamMember, TeamPayout
from discord_self._vendor.discord.template import Template
from discord_self._vendor.discord.threads import Thread, ThreadMember
from discord_self._vendor.discord.tracking import ContextProperties
from discord_self._vendor.discord.user import ClientUser, Note, User
from discord_self._vendor.discord.utils import parse_time
from discord_self._vendor.discord.voice_client import VoiceClient, VoiceProtocol
from discord_self._vendor.discord.webhook.async_ import (
    PartialWebhookChannel,
    PartialWebhookGuild,
    Webhook,
    WebhookMessage,
)
from discord_self._vendor.discord.webhook.sync import SyncWebhook, SyncWebhookMessage
from discord_self._vendor.discord.welcome_screen import WelcomeChannel, WelcomeScreen
from discord_self._vendor.discord.widget import Widget, WidgetChannel, WidgetMember

__all__ = [
    "Achievement",
    "ActionRow",
    "Activity",
    "ActivityType",
    "AllowedMentions",
    "AppCommandOptionType",
    "AppCommandType",
    "Application",
    "ApplicationActivityStatistics",
    "ApplicationAsset",
    "ApplicationAssetType",
    "ApplicationBot",
    "ApplicationBranch",
    "ApplicationBuild",
    "ApplicationCommandCounts",
    "ApplicationDiscoverabilityState",
    "ApplicationDiscoveryFlags",
    "ApplicationExecutable",
    "ApplicationFlags",
    "ApplicationInstallParams",
    "ApplicationMembershipState",
    "ApplicationProfile",
    "ApplicationTester",
    "ApplicationType",
    "ApplicationVerificationState",
    "Asset",
    "Attachment",
    "AudioContext",
    "AudioSource",
    "AuditLogAction",
    "AuditLogActionCategory",
    "AuditLogChanges",
    "AuditLogDiff",
    "AuditLogEntry",
    "AuthFailure",
    "AutoModAction",
    "AutoModPresets",
    "AutoModRule",
    "AutoModRuleAction",
    "AutoModRuleActionType",
    "AutoModRuleEventType",
    "AutoModRuleTriggerType",
    "AutoModTrigger",
    "BanEntry",
    "BaseActivity",
    "BaseCommand",
    "BillingAddress",
    "BotIntegration",
    "Button",
    "ButtonStyle",
    "CallMessage",
    "Capabilities",
    "CaptchaHandler",
    "CaptchaRequired",
    "CategoryChannel",
    "ChannelFlags",
    "ChannelSettings",
    "ChannelType",
    "Client",
    "ClientException",
    "ClientType",
    "ClientUser",
    "Color",
    "Colour",
    "Company",
    "Component",
    "ComponentType",
    "Connection",
    "ConnectionClosed",
    "ConnectionType",
    "ContentFilter",
    "ContentRating",
    "ContentRatingAgency",
    "ContextProperties",
    "CustomActivity",
    "DMChannel",
    "DefaultAvatar",
    "DeletedReferencedMessage",
    "DiscordException",
    "DiscordServerError",
    "Distributor",
    "EULA",
    "EmailSettings",
    "Embed",
    "EmbeddedActivityConfig",
    "EmbeddedActivityOrientation",
    "EmbeddedActivityPlatform",
    "Emoji",
    "EmojiPickerSection",
    "Entitlement",
    "EntitlementPayment",
    "EntitlementType",
    "EntityType",
    "EventStatus",
    "ExpireBehavior",
    "ExpireBehaviour",
    "FFmpegAudio",
    "FFmpegOpusAudio",
    "FFmpegPCMAudio",
    "File",
    "Forbidden",
    "ForumChannel",
    "ForumLayoutType",
    "ForumOrderType",
    "ForumTag",
    "FriendDiscoveryFlags",
    "FriendSourceFlags",
    "Game",
    "GatewayNotFound",
    "Gift",
    "GiftBatch",
    "GiftFlags",
    "GiftStyle",
    "GroupCall",
    "GroupChannel",
    "Guild",
    "GuildAffinity",
    "GuildFolder",
    "GuildProgress",
    "GuildSettings",
    "GuildSticker",
    "HTTPException",
    "HighlightLevel",
    "HubProgressFlags",
    "HypeSquadHouse",
    "InboxTab",
    "Integration",
    "IntegrationAccount",
    "IntegrationApplication",
    "Interaction",
    "InteractionType",
    "InvalidData",
    "Invite",
    "InviteTarget",
    "InviteType",
    "LegacyUserSettings",
    "LibraryApplication",
    "LibraryApplicationFlags",
    "LibrarySKU",
    "Locale",
    "LoginFailure",
    "MFALevel",
    "Manifest",
    "ManifestLabel",
    "Member",
    "MemberCacheFlags",
    "MemberFlags",
    "MemberProfile",
    "Message",
    "MessageCommand",
    "MessageFlags",
    "MessageReference",
    "MessageType",
    "Metadata",
    "Modal",
    "MuteConfig",
    "MutualGuild",
    "NSFWLevel",
    "NotFound",
    "Note",
    "NotificationLevel",
    "Object",
    "OnboardingProgressFlags",
    "OperatingSystem",
    "Option",
    "OptionChoice",
    "PCMAudio",
    "PCMVolumeTransformer",
    "PartialApplication",
    "PartialConnection",
    "PartialEmoji",
    "PartialInviteChannel",
    "PartialInviteGuild",
    "PartialMessage",
    "PartialMessageable",
    "PartialWebhookChannel",
    "PartialWebhookGuild",
    "Payment",
    "PaymentFlags",
    "PaymentGateway",
    "PaymentSource",
    "PaymentSourceFlags",
    "PaymentSourceType",
    "PaymentStatus",
    "PayoutAccountStatus",
    "PayoutReportType",
    "PayoutStatus",
    "PermissionOverwrite",
    "Permissions",
    "PremiumGuildSubscription",
    "PremiumGuildSubscriptionCooldown",
    "PremiumGuildSubscriptionSlot",
    "PremiumType",
    "PremiumUsage",
    "PremiumUsageFlags",
    "PricingPromotion",
    "PrivacyLevel",
    "PrivateCall",
    "PrivateUserFlags",
    "Promotion",
    "PromotionFlags",
    "PublicUserFlags",
    "PurchasedFlags",
    "RPCApplicationState",
    "RateLimited",
    "RawBulkMessageDeleteEvent",
    "RawIntegrationDeleteEvent",
    "RawMessageDeleteEvent",
    "RawMessageUpdateEvent",
    "RawReactionActionEvent",
    "RawReactionClearEmojiEvent",
    "RawReactionClearEvent",
    "RawThreadDeleteEvent",
    "RawThreadMembersUpdate",
    "Reaction",
    "Relationship",
    "RelationshipType",
    "ReportType",
    "RequiredActionType",
    "Role",
    "RoleSubscriptionInfo",
    "RoleTags",
    "SKU",
    "SKUAccessLevel",
    "SKUFeature",
    "SKUFlags",
    "SKUGenre",
    "SKUPrice",
    "SKUType",
    "ScheduledEvent",
    "ScheduledEventEntityType",
    "ScheduledEventStatus",
    "SelectMenu",
    "SelectOption",
    "Session",
    "SlashCommand",
    "SpeakingState",
    "SpoilerRenderOptions",
    "Spotify",
    "StageChannel",
    "StageInstance",
    "StandardSticker",
    "Status",
    "Sticker",
    "StickerAnimationOptions",
    "StickerFormatType",
    "StickerItem",
    "StickerPack",
    "StickerPickerSection",
    "StickerType",
    "StoreApplicationState",
    "StoreAsset",
    "StoreListing",
    "StoreNote",
    "StreamIntegration",
    "Streaming",
    "SubCommand",
    "Subscription",
    "SubscriptionDiscount",
    "SubscriptionDiscountType",
    "SubscriptionInterval",
    "SubscriptionInvoice",
    "SubscriptionInvoiceItem",
    "SubscriptionInvoiceStatus",
    "SubscriptionItem",
    "SubscriptionPlan",
    "SubscriptionPlanPrices",
    "SubscriptionPlanPurchaseType",
    "SubscriptionRenewalMutations",
    "SubscriptionStatus",
    "SubscriptionTrial",
    "SubscriptionType",
    "SyncWebhook",
    "SyncWebhookMessage",
    "SystemChannelFlags",
    "SystemRequirements",
    "Team",
    "TeamMember",
    "TeamPayout",
    "Template",
    "TextChannel",
    "TextInput",
    "TextStyle",
    "Theme",
    "ThirdPartySKU",
    "Thread",
    "ThreadMember",
    "TrackingSettings",
    "TrialOffer",
    "User",
    "UserAffinity",
    "UserCommand",
    "UserContentFilter",
    "UserFlags",
    "UserGuild",
    "UserProfile",
    "UserSettings",
    "VerificationLevel",
    "VideoQualityMode",
    "VoiceChannel",
    "VoiceClient",
    "VoiceProtocol",
    "VoiceState",
    "Webhook",
    "WebhookMessage",
    "WebhookType",
    "WelcomeChannel",
    "WelcomeScreen",
    "Widget",
    "WidgetChannel",
    "WidgetMember",
    "__author__",
    "__copyright__",
    "__license__",
    "__path__",
    "__title__",
    "__version__",
    "abc",
    "opus",
    "parse_time",
    "utils",
    "version_info",
]



from discord_self._vendor.discord.integrations import (
    BotIntegration,
    Integration,
    IntegrationAccount,
    StreamIntegration,
)

__all__ = ["BotIntegration", "Integration", "IntegrationAccount", "StreamIntegration"]



from discord_self._vendor.discord.types.message import Message
from discord_self._vendor.discord.types.snowflake import Snowflake
from discord_self._vendor.discord.types.threads import (
    ForumThread,
    Thread,
    ThreadArchiveDuration,
    ThreadMember,
    ThreadMetadata,
    ThreadPaginationPayload,
    ThreadType,
)

__all__ = [
    "ForumThread",
    "Message",
    "Snowflake",
    "Thread",
    "ThreadArchiveDuration",
    "ThreadMember",
    "ThreadMetadata",
    "ThreadPaginationPayload",
    "ThreadType",
]



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import io
import os
from base64 import b64encode
from hashlib import md5
from typing import Any, Dict, Optional, Tuple, Union

from .utils import MISSING, cached_slot_property

# fmt: off
__all__ = (
    'File',
)
# fmt: on


def _strip_spoiler(filename: str) -> Tuple[str, bool]:
    stripped = filename
    while stripped.startswith("SPOILER_"):
        stripped = stripped[8:]  # len('SPOILER_')
    spoiler = stripped != filename
    return stripped, spoiler


class File:
    r"""A parameter object used for :meth:`abc.Messageable.send`
    for sending file objects.

    .. note::

        File objects are single use and are not meant to be reused in
        multiple :meth:`abc.Messageable.send`\s.

    Attributes
    -----------
    fp: Union[:class:`os.PathLike`, :class:`io.BufferedIOBase`]
        A file-like object opened in binary mode and read mode
        or a filename representing a file in the hard drive to
        open.

        .. note::

            If the file-like object passed is opened via ``open`` then the
            modes 'rb' should be used.

            To pass binary data, consider usage of ``io.BytesIO``.

    spoiler: :class:`bool`
        Whether the attachment is a spoiler. If left unspecified, the :attr:`~File.filename` is used
        to determine if the file is a spoiler.
    description: Optional[:class:`str`]
        The file description to display, currently only supported for images.

        .. versionadded:: 2.0
    """

    __slots__ = (
        "fp",
        "_filename",
        "spoiler",
        "description",
        "_original_pos",
        "_owner",
        "_closer",
        "_cs_md5",
    )

    def __init__(
        self,
        fp: Union[str, bytes, os.PathLike[Any], io.BufferedIOBase],
        filename: Optional[str] = None,
        *,
        spoiler: bool = MISSING,
        description: Optional[str] = None,
    ):
        if isinstance(fp, io.IOBase):
            if not (fp.seekable() and fp.readable()):
                raise ValueError(f"File buffer {fp!r} must be seekable and readable")
            self.fp: io.BufferedIOBase = fp
            self._original_pos = fp.tell()
            self._owner = False
        else:
            self.fp = open(fp, "rb")
            self._original_pos = 0
            self._owner = True

        # aiohttp only uses two methods from IOBase (read and close)
        # Since I want to control when the files close,
        # I need to stub it so it doesn't close unless I tell it to
        self._closer = self.fp.close
        self.fp.close = lambda: None

        if filename is None:
            if isinstance(fp, str):
                _, filename = os.path.split(fp)
            else:
                filename = getattr(fp, "name", "untitled")

        self._filename, filename_spoiler = _strip_spoiler(filename)
        if spoiler is MISSING:
            spoiler = filename_spoiler

        self.spoiler: bool = spoiler
        self.description: Optional[str] = description

    @property
    def filename(self) -> str:
        """:class:`str`: The filename to display when uploading to Discord.
        If this is not given then it defaults to ``fp.name`` or if ``fp`` is
        a string then the ``filename`` will default to the string given.
        """
        return "SPOILER_" + self._filename if self.spoiler else self._filename

    @filename.setter
    def filename(self, value: str) -> None:
        self._filename, self.spoiler = _strip_spoiler(value)

    @cached_slot_property("_cs_md5")
    def md5(self) -> str:
        try:
            return b64encode(md5(self.fp.read()).digest()).decode("utf-8")
        finally:
            self.reset()

    def reset(self, *, seek: Union[int, bool] = True) -> None:
        # The `seek` parameter is needed because
        # the retry-loop is iterated over multiple times
        # starting from 0, as an implementation quirk
        # the resetting must be done at the beginning
        # before a request is done, since the first index
        # is 0, and thus false, then this prevents an
        # unnecessary seek since it's the first request
        # done.
        if seek:
            self.fp.seek(self._original_pos)

    def close(self) -> None:
        self.fp.close = self._closer
        if self._owner:
            self._closer()

    def to_dict(self, index: int) -> Dict[str, Any]:
        payload = {
            "id": index,
            "filename": self.filename,
        }

        if self.description is not None:
            payload["description"] = self.description

        return payload



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Iterator, Optional, Tuple, Union

from .utils import parse_time


class Metadata:
    """Represents a raw model from Discord.

    Because of how unstable and wildly varying some metadata in Discord can be, this is a simple class
    that just provides access to the raw data using dot notation. This means that ``None`` is returned
    for unknown attributes instead of raising an exception. This class can be used similarly to a dictionary.

    .. versionadded:: 2.0

    .. container:: operations

        .. describe:: x == y

            Checks if two metadata objects are equal.

        .. describe:: x != y

            Checks if two metadata objects are not equal.

        .. describe:: x[key]

            Returns a metadata value if it is found, otherwise raises a :exc:`KeyError`.

        .. describe:: key in x

            Checks if a metadata value is present.

        .. describe:: len(x)

            Returns the number of metadata values present.

        .. describe:: iter(x)
            Returns an iterator of ``(field, value)`` pairs. This allows this class
            to be used as an iterable in list/dict/etc constructions.
    """

    def __init__(self, data: Optional[MetadataObject] = None) -> None:
        if not data:
            return

        for key, value in data.items():
            if isinstance(value, dict):
                value = Metadata(value)
            elif key.endswith("_id") and isinstance(value, str) and value.isdigit():
                value = int(value)
            elif (key.endswith("_at") or key.endswith("_date")) and isinstance(
                value, str
            ):
                try:
                    value = parse_time(value)
                except ValueError:
                    pass
            elif isinstance(value, list):
                value = [Metadata(x) if isinstance(x, dict) else x for x in value]

            self.__dict__[key] = value

    def __repr__(self) -> str:
        if not self.__dict__:
            return "<Metadata>"
        return f"<Metadata {' '.join(f'{k}={v!r}' for k, v in self.__dict__.items())}>"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Metadata):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, Metadata):
            return True
        return self.__dict__ != other.__dict__

    def __iter__(self) -> Iterator[Tuple[str, Any]]:
        yield from self.__dict__.items()

    def __getitem__(self, key: str) -> Any:
        return self.__dict__[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.__dict__[key] = value

    def __getattr__(self, _) -> Any:
        return None

    def __contains__(self, key: str) -> bool:
        return key in self.__dict__

    def __len__(self) -> int:
        return len(self.__dict__)

    def keys(self):
        """A set-like object providing a view on the metadata's keys."""
        return self.__dict__.keys()

    def values(self):
        """A set-like object providing a view on the metadata's values."""
        return self.__dict__.values()

    def items(self):
        """A set-like object providing a view on the metadata's items."""
        return self.__dict__.items()


if TYPE_CHECKING:
    MetadataObject = Union[Metadata, Dict[str, Any]]



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import array
import asyncio
import copy
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Sequence,
    Union,
)

from .abc import GuildChannel, Messageable, _purge_helper
from .enums import ChannelType, try_enum
from .errors import ClientException, InvalidData
from .flags import ChannelFlags
from .mixins import Hashable
from .permissions import Permissions
from .utils import MISSING, _get_as_snowflake, _unique, parse_time, snowflake_time

__all__ = (
    "Thread",
    "ThreadMember",
)

if TYPE_CHECKING:
    from datetime import datetime

    from typing_extensions import Self

    from .abc import Snowflake, SnowflakeTime
    from .channel import CategoryChannel, ForumChannel, ForumTag, TextChannel
    from .guild import Guild
    from .member import Member
    from .message import Message, PartialMessage
    from .role import Role
    from .state import ConnectionState
    from .types.threads import (
        Thread as ThreadPayload,
    )
    from .types.threads import (
        ThreadArchiveDuration,
        ThreadMetadata,
    )
    from .types.threads import (
        ThreadMember as ThreadMemberPayload,
    )


class Thread(Messageable, Hashable):
    """Represents a Discord thread.

    .. container:: operations

        .. describe:: x == y

            Checks if two threads are equal.

        .. describe:: x != y

            Checks if two threads are not equal.

        .. describe:: hash(x)

            Returns the thread's hash.

        .. describe:: str(x)

            Returns the thread's name.

    .. versionadded:: 2.0

    Attributes
    -----------
    name: :class:`str`
        The thread name.
    guild: :class:`Guild`
        The guild the thread belongs to.
    id: :class:`int`
        The thread ID. This is the same as the thread starter message ID.
    parent_id: :class:`int`
        The ID of the parent :class:`TextChannel` or :class:`ForumChannel` this thread belongs to.
    owner_id: :class:`int`
        The ID of the user that created this thread.
    last_message_id: Optional[:class:`int`]
        The last message ID of the message sent to this thread. It may
        *not* point to an existing or valid message.
    last_pin_timestamp: Optional[:class:`datetime.datetime`]
        When the last pinned message was pinned. ``None`` if there are no pinned messages.
    slowmode_delay: :class:`int`
        The number of seconds a member must wait between sending messages
        in this thread. A value of ``0`` denotes that it is disabled.
        Bots and users with :attr:`~Permissions.manage_channels` or
        :attr:`~Permissions.manage_messages` bypass slowmode.
    message_count: :class:`int`
        An approximate number of messages in this thread.
    member_count: :class:`int`
        An approximate number of members in this thread. This caps at 50.
    archived: :class:`bool`
        Whether the thread is archived.
    locked: :class:`bool`
        Whether the thread is locked.
    invitable: :class:`bool`
        Whether non-moderators can add other non-moderators to this thread.
        This is always ``True`` for public threads.
    auto_archive_duration: :class:`int`
        The duration in minutes until the thread is automatically archived due to inactivity.
        Usually a value of 60, 1440, 4320 and 10080.
    archive_timestamp: :class:`datetime.datetime`
        An aware timestamp of when the thread's archived status was last updated in UTC.
    """

    __slots__ = (
        "name",
        "id",
        "guild",
        "_type",
        "_state",
        "_members",
        "owner_id",
        "parent_id",
        "last_message_id",
        "last_pin_timestamp",
        "message_count",
        "member_count",
        "slowmode_delay",
        "locked",
        "archived",
        "invitable",
        "auto_archive_duration",
        "archive_timestamp",
        "_member_ids",
        "_created_at",
        "_flags",
        "_applied_tags",
    )

    def __init__(
        self, *, guild: Guild, state: ConnectionState, data: ThreadPayload
    ) -> None:
        self._state: ConnectionState = state
        self.guild: Guild = guild
        self._members: Dict[int, ThreadMember] = {}
        self._from_data(data)

    async def _get_channel(self) -> Self:
        return self

    def __repr__(self) -> str:
        return (
            f"<Thread id={self.id!r} name={self.name!r} parent={self.parent}"
            f" owner_id={self.owner_id!r} locked={self.locked} archived={self.archived}>"
        )

    def __str__(self) -> str:
        return self.name

    def _from_data(self, data: ThreadPayload):
        self.id: int = int(data["id"])
        self.parent_id: int = int(data["parent_id"])
        self.owner_id: int = int(data["owner_id"])
        self.name: str = data["name"]
        self._type: ChannelType = try_enum(ChannelType, data["type"])
        self.last_message_id: Optional[int] = _get_as_snowflake(data, "last_message_id")
        self.last_pin_timestamp: Optional[datetime] = parse_time(
            data.get("last_pin_timestamp")
        )
        self.slowmode_delay: int = data.get("rate_limit_per_user", 0)
        self.message_count: int = data["message_count"]
        self.member_count: int = data["member_count"]
        self._member_ids: List[Union[str, int]] = data["member_ids_preview"]
        self._flags: int = data.get("flags", 0)
        # SnowflakeList is sorted, but this would not be proper for applied tags, where order actually matters.
        self._applied_tags: array.array[int] = array.array(
            "Q", map(int, data.get("applied_tags", []))
        )
        self._unroll_metadata(data["thread_metadata"])

    def _unroll_metadata(self, data: ThreadMetadata):
        self.archived: bool = data["archived"]
        self.auto_archive_duration: int = data["auto_archive_duration"]
        self.archive_timestamp: datetime = parse_time(data["archive_timestamp"])
        self.locked: bool = data.get("locked", False)
        self.invitable: bool = data.get("invitable", True)
        self._created_at: Optional[datetime] = parse_time(data.get("create_timestamp"))

    def _update(self, data):
        old = copy.copy(self)
        self.slowmode_delay = data.get("rate_limit_per_user", 0)
        self._flags: int = data.get("flags", 0)
        self._applied_tags: array.array[int] = array.array(
            "Q", map(int, data.get("applied_tags", []))
        )

        if (meta := data.get("thread_metadata")) is not None:
            self._unroll_metadata(meta)
        if (name := data.get("name")) is not None:
            self.name = name
        if (last_message_id := _get_as_snowflake(data, "last_message_id")) is not None:
            self.last_message_id = last_message_id
        if (message_count := data.get("message_count")) is not None:
            self.message_count = message_count
        if (member_count := data.get("member_count")) is not None:
            self.member_count = member_count
        if (member_ids := data.get("member_ids_preview")) is not None:
            self._member_ids = member_ids

        attrs = [
            x
            for x in self.__slots__
            if not any(y in x for y in ("member", "guild", "state", "count"))
        ]

        if any(getattr(self, attr) != getattr(old, attr) for attr in attrs):
            return old

    @property
    def type(self) -> ChannelType:
        """:class:`ChannelType`: The channel's Discord type."""
        return self._type

    @property
    def parent(self) -> Optional[Union[TextChannel, ForumChannel]]:
        """Optional[Union[:class:`ForumChannel`, :class:`TextChannel`]]: The parent channel this thread belongs to.

        There is an alias for this named :attr:`channel`.
        """
        return self.guild.get_channel(self.parent_id)  # type: ignore

    @property
    def channel(self) -> Optional[Union[TextChannel, ForumChannel]]:
        """Optional[Union[:class:`ForumChannel`, :class:`TextChannel`]]: The parent channel this thread belongs to.

        This is an alias of :attr:`parent`.
        """
        return self.parent

    @property
    def flags(self) -> ChannelFlags:
        """:class:`ChannelFlags`: The flags associated with this thread."""
        return ChannelFlags._from_value(self._flags)

    @property
    def owner(self) -> Optional[Member]:
        """Optional[:class:`Member`]: The member this thread belongs to."""
        return self.guild.get_member(self.owner_id)

    @property
    def mention(self) -> str:
        """:class:`str`: The string that allows you to mention the thread."""
        return f"<#{self.id}>"

    @property
    def created_at(self) -> datetime:
        """:class:`datetime.datetime`: Returns the thread's creation time in UTC.

        .. note::

            This may be inaccurate for threads created before January 9th, 2022.
        """
        return self._created_at or snowflake_time(self.id)

    @property
    def jump_url(self) -> str:
        """:class:`str`: Returns a URL that allows the client to jump to the thread."""
        return f"https://discord.com/channels/{self.guild.id}/{self.id}"

    @property
    def members(self) -> List[ThreadMember]:
        """List[:class:`ThreadMember`]: A list of thread members in this thread.

        Initial members are not provided by Discord. You must call :func:`fetch_members`.
        """
        return list(self._members.values())

    @property
    def applied_tags(self) -> List[ForumTag]:
        """List[:class:`ForumTag`]: A list of tags applied to this thread."""
        tags = []
        if self.parent is None or self.parent.type != ChannelType.forum:
            return tags

        parent = self.parent
        for tag_id in self._applied_tags:
            tag = parent.get_tag(tag_id)  # type: ignore
            if tag is not None:
                tags.append(tag)

        return tags

    @property
    def starter_message(self) -> Optional[Message]:
        """Returns the thread starter message from the cache.

        The message might not be cached, valid, or point to an existing message.

        Note that the thread starter message ID is the same ID as the thread.

        Returns
        --------
        Optional[:class:`Message`]
            The thread starter message or ``None`` if not found.
        """
        return self._state._get_message(self.id)

    @property
    def last_message(self) -> Optional[Message]:
        """Returns the last message from this thread from the cache.

        The message might not be valid or point to an existing message.

        .. admonition:: Reliable Fetching
            :class: helpful

            For a slightly more reliable method of fetching the
            last message, consider using either :meth:`history`
            or :meth:`fetch_message` with the :attr:`last_message_id`
            attribute.

        Returns
        ---------
        Optional[:class:`Message`]
            The last message in this channel or ``None`` if not found.
        """
        return (
            self._state._get_message(self.last_message_id)
            if self.last_message_id
            else None
        )

    @property
    def category(self) -> Optional[CategoryChannel]:
        """The category channel the parent channel belongs to, if applicable.

        Raises
        -------
        ClientException
            The parent channel was not cached and returned ``None``.

        Returns
        -------
        Optional[:class:`CategoryChannel`]
            The parent channel's category.
        """

        parent = self.parent
        if parent is None:
            raise ClientException("Parent channel not found")
        return parent.category

    @property
    def category_id(self) -> Optional[int]:
        """The category channel ID the parent channel belongs to, if applicable.

        Raises
        -------
        ClientException
            The parent channel was not cached and returned ``None``.

        Returns
        -------
        Optional[:class:`int`]
            The parent channel's category ID.
        """

        parent = self.parent
        if parent is None:
            raise ClientException("Parent channel not found")
        return parent.category_id

    @property
    def me(self) -> Optional[ThreadMember]:
        """Optional[:class:`ThreadMember`]: A thread member representing yourself, if you've joined the thread.

        This might not be available.
        """
        self_id = self._state.self_id
        return self._members.get(self_id)  # type: ignore

    @me.setter
    def me(self, member) -> None:
        self._members[member.id] = member

    def is_private(self) -> bool:
        """:class:`bool`: Whether the thread is a private thread.

        A private thread is only viewable by those that have been explicitly
        invited or have :attr:`~.Permissions.manage_threads`.
        """
        return self._type is ChannelType.private_thread

    def is_news(self) -> bool:
        """:class:`bool`: Whether the thread is a news thread.

        A news thread is a thread that has a parent that is a news channel,
        i.e. :meth:`.TextChannel.is_news` is ``True``.
        """
        return self._type is ChannelType.news_thread

    def is_nsfw(self) -> bool:
        """:class:`bool`: Whether the thread is NSFW or not.

        An NSFW thread is a thread that has a parent that is an NSFW channel,
        i.e. :meth:`.TextChannel.is_nsfw` is ``True``.
        """
        parent = self.parent
        return parent is not None and parent.is_nsfw()

    def permissions_for(self, obj: Union[Member, Role], /) -> Permissions:
        """Handles permission resolution for the :class:`~discord.Member`
        or :class:`~discord.Role`.

        Since threads do not have their own permissions, they mostly
        inherit them from the parent channel with some implicit
        permissions changed.

        Parameters
        ----------
        obj: Union[:class:`~discord.Member`, :class:`~discord.Role`]
            The object to resolve permissions for. This could be either
            a member or a role. If it's a role then member overwrites
            are not computed.

        Raises
        -------
        ClientException
            The parent channel was not cached and returned ``None``

        Returns
        -------
        :class:`~discord.Permissions`
            The resolved permissions for the member or role.
        """

        parent = self.parent
        if parent is None:
            raise ClientException("Parent channel not found")

        base = GuildChannel.permissions_for(parent, obj)

        # if you can't send a message in a channel then you can't have certain
        # permissions as well
        if not base.send_messages_in_threads:
            base.send_tts_messages = False
            base.mention_everyone = False
            base.embed_links = False
            base.attach_files = False

        # if you can't read a channel then you have no permissions there
        if not base.read_messages:
            denied = Permissions.all_channel()
            base.value &= ~denied.value

        return base

    async def delete_messages(
        self, messages: Iterable[Snowflake], /, *, reason: Optional[str] = None
    ) -> None:
        """|coro|

        Deletes a list of messages. This is similar to :meth:`Message.delete`
        except it bulk deletes multiple messages.

        You must have :attr:`~Permissions.manage_messages` to use this (unless they're your own).

        .. note::
            Users do not have access to the message bulk-delete endpoint.
            Since messages are just iterated over and deleted one-by-one,
            it's easy to get ratelimited using this method.

        Parameters
        -----------
        messages: Iterable[:class:`abc.Snowflake`]
            An iterable of messages denoting which ones to bulk delete.
        reason: Optional[:class:`str`]
            The reason for deleting the messages. Shows up on the audit log.

        Raises
        ------
        Forbidden
            You do not have proper permissions to delete the messages.
        HTTPException
            Deleting the messages failed.
        """
        if not isinstance(messages, (list, tuple)):
            messages = list(messages)

        if len(messages) == 0:
            return  # Do nothing

        await self._state._delete_messages(self.id, messages, reason=reason)

    async def purge(
        self,
        *,
        limit: Optional[int] = 100,
        check: Callable[[Message], bool] = MISSING,
        before: Optional[SnowflakeTime] = None,
        after: Optional[SnowflakeTime] = None,
        around: Optional[SnowflakeTime] = None,
        oldest_first: Optional[bool] = None,
        reason: Optional[str] = None,
    ) -> List[Message]:
        """|coro|

        Purges a list of messages that meet the criteria given by the predicate
        ``check``. If a ``check`` is not provided then all messages are deleted
        without discrimination.

        Having :attr:`~Permissions.read_message_history` is needed to
        retrieve message history.

        Examples
        ---------

        Deleting bot's messages ::

            def is_me(m):
                return m.author == client.user

            deleted = await channel.purge(limit=100, check=is_me)
            await channel.send(f'Deleted {len(deleted)} message(s)')

        Parameters
        -----------
        limit: Optional[:class:`int`]
            The number of messages to search through. This is not the number
            of messages that will be deleted, though it can be.
        check: Callable[[:class:`Message`], :class:`bool`]
            The function used to check if a message should be deleted.
            It must take a :class:`Message` as its sole parameter.
        before: Optional[Union[:class:`abc.Snowflake`, :class:`datetime.datetime`]]
            Same as ``before`` in :meth:`history`.
        after: Optional[Union[:class:`abc.Snowflake`, :class:`datetime.datetime`]]
            Same as ``after`` in :meth:`history`.
        around: Optional[Union[:class:`abc.Snowflake`, :class:`datetime.datetime`]]
            Same as ``around`` in :meth:`history`.
        oldest_first: Optional[:class:`bool`]
            Same as ``oldest_first`` in :meth:`history`.
        reason: Optional[:class:`str`]
            The reason for purging the messages. Shows up on the audit log.

        Raises
        -------
        Forbidden
            You do not have proper permissions to do the actions required.
        HTTPException
            Purging the messages failed.

        Returns
        --------
        List[:class:`.Message`]
            The list of messages that were deleted.
        """
        return await _purge_helper(
            self,
            limit=limit,
            check=check,
            before=before,
            after=after,
            around=around,
            oldest_first=oldest_first,
            reason=reason,
        )

    async def edit(
        self,
        *,
        name: str = MISSING,
        archived: bool = MISSING,
        locked: bool = MISSING,
        invitable: bool = MISSING,
        pinned: bool = MISSING,
        slowmode_delay: int = MISSING,
        auto_archive_duration: ThreadArchiveDuration = MISSING,
        applied_tags: Sequence[ForumTag] = MISSING,
        reason: Optional[str] = None,
    ) -> Thread:
        """|coro|

        Edits the thread.

        Editing the thread requires :attr:`.Permissions.manage_threads`. The thread
        creator can also edit ``name``, ``archived`` or ``auto_archive_duration``.
        Note that if the thread is locked then only those with :attr:`.Permissions.manage_threads`
        can unarchive a thread.

        The thread must be unarchived to be edited.

        Parameters
        ------------
        name: :class:`str`
            The new name of the thread.
        archived: :class:`bool`
            Whether to archive the thread or not.
        locked: :class:`bool`
            Whether to lock the thread or not.
        pinned: :class:`bool`
            Whether to pin the thread or not. This only works if the thread is part of a forum.
        invitable: :class:`bool`
            Whether non-moderators can add other non-moderators to this thread.
            Only available for private threads.
        auto_archive_duration: :class:`int`
            The new duration in minutes before a thread is automatically archived for inactivity.
            Must be one of ``60``, ``1440``, ``4320``, or ``10080``.
        slowmode_delay: :class:`int`
            Specifies the slowmode rate limit for user in this thread, in seconds.
            A value of ``0`` disables slowmode. The maximum value possible is ``21600``.
        applied_tags: Sequence[:class:`ForumTag`]
            The new tags to apply to the thread. There can only be up to 5 tags applied to a thread.
        reason: Optional[:class:`str`]
            The reason for editing this thread. Shows up on the audit log.

        Raises
        -------
        Forbidden
            You do not have permissions to edit the thread.
        HTTPException
            Editing the thread failed.

        Returns
        --------
        :class:`Thread`
            The newly edited thread.
        """
        payload = {}
        if name is not MISSING:
            payload["name"] = str(name)
        if archived is not MISSING:
            payload["archived"] = archived
        if auto_archive_duration is not MISSING:
            payload["auto_archive_duration"] = auto_archive_duration
        if locked is not MISSING:
            payload["locked"] = locked
        if invitable is not MISSING:
            payload["invitable"] = invitable
        if slowmode_delay is not MISSING:
            payload["rate_limit_per_user"] = slowmode_delay
        if pinned is not MISSING:
            flags = self.flags
            flags.pinned = pinned
            payload["flags"] = flags.value
        if applied_tags is not MISSING:
            payload["applied_tags"] = [str(tag.id) for tag in applied_tags]

        data = await self._state.http.edit_channel(self.id, **payload, reason=reason)
        # The data payload will always be a Thread payload
        return Thread(data=data, state=self._state, guild=self.guild)  # type: ignore

    async def add_tags(self, *tags: Snowflake, reason: Optional[str] = None) -> None:
        r"""|coro|

        Adds the given forum tags to a thread.

        You must have :attr:`~Permissions.manage_threads` to
        use this or the thread must be owned by you.

        Tags that have :attr:`ForumTag.moderated` set to ``True`` require
        :attr:`~Permissions.manage_threads` to be added.

        The maximum number of tags that can be added to a thread is 5.

        The parent channel must be a :class:`ForumChannel`.

        Parameters
        -----------
        \*tags: :class:`abc.Snowflake`
            An argument list of :class:`abc.Snowflake` representing a :class:`ForumTag`
            to add to the thread.
        reason: Optional[:class:`str`]
            The reason for adding these tags.

        Raises
        -------
        Forbidden
            You do not have permissions to add these tags.
        HTTPException
            Adding tags failed.
        """

        applied_tags = [str(tag) for tag in self._applied_tags]
        applied_tags.extend(str(tag.id) for tag in tags)

        await self._state.http.edit_channel(
            self.id, applied_tags=_unique(applied_tags), reason=reason
        )

    async def remove_tags(self, *tags: Snowflake, reason: Optional[str] = None) -> None:
        r"""|coro|

        Remove the given forum tags to a thread.

        You must have :attr:`~Permissions.manage_threads` to
        use this or the thread must be owned by you.

        The parent channel must be a :class:`ForumChannel`.

        Parameters
        -----------
        \*tags: :class:`abc.Snowflake`
            An argument list of :class:`abc.Snowflake` representing a :class:`ForumTag`
            to remove to the thread.
        reason: Optional[:class:`str`]
            The reason for removing these tags.

        Raises
        -------
        Forbidden
            You do not have permissions to remove these tags.
        HTTPException
            Removing tags failed.
        """

        # Once again, taking advantage of the fact that dicts are ordered since 3.7
        applied_tags: Dict[str, Literal[None]] = {
            str(tag): None for tag in self._applied_tags
        }

        for tag in tags:
            applied_tags.pop(str(tag.id), None)

        await self._state.http.edit_channel(
            self.id, applied_tags=list(applied_tags.keys()), reason=reason
        )

    async def join(self) -> None:
        """|coro|

        Joins this thread.

        You must have :attr:`~Permissions.send_messages_in_threads` to join a thread.
        If the thread is private, :attr:`~Permissions.manage_threads` is also needed.

        Raises
        -------
        Forbidden
            You do not have permissions to join the thread.
        HTTPException
            Joining the thread failed.
        """
        await self._state.http.join_thread(self.id)

    async def leave(self) -> None:
        """|coro|

        Leaves this thread.

        Raises
        -------
        HTTPException
            Leaving the thread failed.
        """
        await self._state.http.leave_thread(self.id)

    async def add_user(self, user: Snowflake, /) -> None:
        """|coro|

        Adds a user to this thread.

        You must have :attr:`~Permissions.send_messages_in_threads` to add a user to a thread.
        If the thread is private and :attr:`invitable` is ``False`` then :attr:`~Permissions.manage_messages`
        is required to add a user to the thread.

        Parameters
        -----------
        user: :class:`abc.Snowflake`
            The user to add to the thread.

        Raises
        -------
        Forbidden
            You do not have permissions to add the user to the thread.
        HTTPException
            Adding the user to the thread failed.
        """
        await self._state.http.add_user_to_thread(self.id, user.id)

    async def remove_user(self, user: Snowflake, /) -> None:
        """|coro|

        Removes a user from this thread.

        You must have :attr:`~Permissions.manage_threads` or be the creator of the thread to remove a user.

        Parameters
        -----------
        user: :class:`abc.Snowflake`
            The user to remove from the thread.

        Raises
        -------
        Forbidden
            You do not have permissions to remove the user from the thread.
        HTTPException
            Removing the user from the thread failed.
        """
        await self._state.http.remove_user_from_thread(self.id, user.id)

    async def fetch_members(self) -> List[ThreadMember]:
        """|coro|

        Retrieves all :class:`ThreadMember` that are in this thread,
        along with their respective :class:`Member`.

        Raises
        -------
        InvalidData
            Discord didn't respond with the members.

        Returns
        --------
        List[:class:`ThreadMember`]
            All thread members in the thread.
        """
        state = self._state
        await state.ws.request_lazy_guild(
            self.parent.guild.id, thread_member_lists=[self.id]
        )  # type: ignore
        future = state.ws.wait_for(
            "thread_member_list_update", lambda d: int(d["thread_id"]) == self.id
        )
        try:
            data = await asyncio.wait_for(future, timeout=15)
        except asyncio.TimeoutError as exc:
            raise InvalidData("Didn't receieve a response from Discord") from exc

        members = [ThreadMember(self, {"member": member}) for member in data["members"]]  # type: ignore
        for m in members:
            self._add_member(m)

        return self.members  # Includes correct self.me

    async def delete(self) -> None:
        """|coro|

        Deletes this thread.

        You must have :attr:`~Permissions.manage_threads` to delete threads.

        Raises
        -------
        Forbidden
            You do not have permissions to delete this thread.
        HTTPException
            Deleting the thread failed.
        """
        await self._state.http.delete_channel(self.id)

    def get_partial_message(self, message_id: int, /) -> PartialMessage:
        """Creates a :class:`PartialMessage` from the message ID.

        This is useful if you want to work with a message and only have its ID without
        doing an unnecessary API call.

        Parameters
        ------------
        message_id: :class:`int`
            The message ID to create a partial message for.

        Returns
        ---------
        :class:`PartialMessage`
            The partial message.
        """
        from .message import PartialMessage

        return PartialMessage(channel=self, id=message_id)

    def _add_member(self, member: ThreadMember, /) -> None:
        if member.id != self._state.self_id:
            self._members[member.id] = member

    def _pop_member(self, member_id: int, /) -> Optional[ThreadMember]:
        return self._members.pop(member_id, None)


class ThreadMember(Hashable):
    """Represents a Discord thread member.

    .. container:: operations

        .. describe:: x == y

            Checks if two thread members are equal.

        .. describe:: x != y

            Checks if two thread members are not equal.

        .. describe:: hash(x)

            Returns the thread member's hash.

        .. describe:: str(x)

            Returns the thread member's name.

    .. versionadded:: 2.0

    Attributes
    -----------
    id: :class:`int`
        The thread member's ID.
    thread_id: :class:`int`
        The thread's ID.
    joined_at: Optional[:class:`datetime.datetime`]
        The time the member joined the thread in UTC.
        Only reliably available for yourself or members joined while the user is connected to the gateway.
    flags: :class:`int`
        The thread member's flags. Will be its own class in the future.
        Only reliably available for yourself or members joined while the user is connected to the gateway.
    """

    __slots__ = (
        "id",
        "thread_id",
        "joined_at",
        "flags",
        "_state",
        "parent",
    )

    def __init__(self, parent: Thread, data: ThreadMemberPayload) -> None:
        self.parent: Thread = parent
        self._state: ConnectionState = parent._state
        self._from_data(data)

    def __repr__(self) -> str:
        return f"<ThreadMember id={self.id} thread_id={self.thread_id} joined_at={self.joined_at!r}>"

    def _from_data(self, data: ThreadMemberPayload) -> None:
        state = self._state

        self.id: int
        try:
            self.id = int(data["user_id"])
        except KeyError:
            self.id = state.self_id  # type: ignore

        self.thread_id: int
        try:
            self.thread_id = int(data["id"])
        except KeyError:
            self.thread_id = self.parent.id

        self.joined_at = parse_time(data.get("join_timestamp"))
        self.flags = data.get("flags")

        if (mdata := data.get("member")) is not None:
            guild = self.parent.guild
            mdata["guild_id"] = guild.id
            self.id = user_id = int(data["user_id"])
            mdata["presence"] = data.get("presence")
            if guild.get_member(user_id) is not None:
                state.parse_guild_member_update(mdata)
            else:
                state.parse_guild_member_add(mdata)

    @property
    def thread(self) -> Thread:
        """:class:`Thread`: The thread this member belongs to."""
        return self.parent

    @property
    def member(self) -> Optional[Member]:
        """Optional[:class:`Member`]: The member this :class:`ThreadMember` represents. If the member
        is not cached then this will be ``None``.
        """
        return self.parent.guild.get_member(self.id)



from discord_self._vendor.discord.types.snowflake import Snowflake
from discord_self._vendor.discord.types.team import Team, TeamMember, TeamPayout
from discord_self._vendor.discord.types.user import PartialUser

__all__ = ["PartialUser", "Snowflake", "Team", "TeamMember", "TeamPayout"]



from discord_self._vendor.discord.gateway import (
    DiscordVoiceWebSocket,
    DiscordWebSocket,
    KeepAliveHandler,
    ReconnectWebSocket,
    VoiceKeepAliveHandler,
)

__all__ = [
    "DiscordVoiceWebSocket",
    "DiscordWebSocket",
    "KeepAliveHandler",
    "ReconnectWebSocket",
    "VoiceKeepAliveHandler",
]



from discord_self._vendor.discord.types.channel import ChannelType
from discord_self._vendor.discord.types.command import (
    ApplicationCommand,
    ApplicationCommandOption,
    ApplicationCommandOptionChoice,
    ApplicationCommandOptionType,
    ApplicationCommandPermissions,
    ApplicationCommandPermissionType,
    ApplicationCommandType,
    GlobalApplicationCommand,
    GuildApplicationCommand,
    GuildApplicationCommandPermissions,
)
from discord_self._vendor.discord.types.snowflake import Snowflake

__all__ = [
    "ApplicationCommand",
    "ApplicationCommandOption",
    "ApplicationCommandOptionChoice",
    "ApplicationCommandOptionType",
    "ApplicationCommandPermissionType",
    "ApplicationCommandPermissions",
    "ApplicationCommandType",
    "ChannelType",
    "GlobalApplicationCommand",
    "GuildApplicationCommand",
    "GuildApplicationCommandPermissions",
    "Snowflake",
]



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
    Union,
    overload,
)

from . import utils
from .enums import (
    AutoModRuleActionType,
    AutoModRuleEventType,
    AutoModRuleTriggerType,
    try_enum,
)
from .flags import AutoModPresets
from .utils import MISSING, cached_slot_property

if TYPE_CHECKING:
    from typing_extensions import Self

    from .abc import GuildChannel, Snowflake
    from .guild import Guild
    from .member import Member
    from .role import Role
    from .state import ConnectionState
    from .threads import Thread
    from .types.automod import (
        AutoModerationAction as AutoModerationActionPayload,
    )
    from .types.automod import (
        AutoModerationActionExecution as AutoModerationActionExecutionPayload,
    )
    from .types.automod import (
        AutoModerationRule as AutoModerationRulePayload,
    )
    from .types.automod import (
        AutoModerationTriggerMetadata as AutoModerationTriggerMetadataPayload,
    )

__all__ = (
    "AutoModRuleAction",
    "AutoModTrigger",
    "AutoModRule",
    "AutoModAction",
)


class AutoModRuleAction:
    """Represents an auto moderation's rule action.

    .. note::
        Only one of ``channel_id``, ``duration``, or ``custom_message`` can be used.

    .. versionadded:: 2.0

    Attributes
    -----------
    type: :class:`AutoModRuleActionType`
        The type of action to take.
        Defaults to :attr:`~AutoModRuleActionType.block_message`.
    channel_id: Optional[:class:`int`]
        The ID of the channel or thread to send the alert message to, if any.
        Passing this sets :attr:`type` to :attr:`~AutoModRuleActionType.send_alert_message`.
    duration: Optional[:class:`datetime.timedelta`]
        The duration of the timeout to apply, if any.
        Has a maximum of 28 days.
        Passing this sets :attr:`type` to :attr:`~AutoModRuleActionType.timeout`.
    custom_message: Optional[:class:`str`]
        A custom message which will be shown to a user when their message is blocked.
        Passing this sets :attr:`type` to :attr:`~AutoModRuleActionType.block_message`.

        .. versionadded:: 2.0
    """

    __slots__ = ("type", "channel_id", "duration", "custom_message")

    @overload
    def __init__(self, *, channel_id: Optional[int] = ...) -> None: ...

    @overload
    def __init__(self, *, duration: Optional[datetime.timedelta] = ...) -> None: ...

    @overload
    def __init__(self, *, custom_message: Optional[str] = ...) -> None: ...

    def __init__(
        self,
        *,
        channel_id: Optional[int] = None,
        duration: Optional[datetime.timedelta] = None,
        custom_message: Optional[str] = None,
    ) -> None:
        self.channel_id: Optional[int] = channel_id
        self.duration: Optional[datetime.timedelta] = duration
        self.custom_message: Optional[str] = custom_message

        if sum(v is None for v in (channel_id, duration, custom_message)) < 2:
            raise ValueError(
                "Only one of channel_id, duration, or custom_message can be passed."
            )

        self.type: AutoModRuleActionType = AutoModRuleActionType.block_message
        if channel_id:
            self.type = AutoModRuleActionType.send_alert_message
        elif duration:
            self.type = AutoModRuleActionType.timeout

    def __repr__(self) -> str:
        return f"<AutoModRuleAction type={self.type.value} channel={self.channel_id} duration={self.duration}>"

    @classmethod
    def from_data(cls, data: AutoModerationActionPayload) -> Self:
        if data["type"] == AutoModRuleActionType.timeout.value:
            duration_seconds = data["metadata"]["duration_seconds"]
            return cls(duration=datetime.timedelta(seconds=duration_seconds))
        elif data["type"] == AutoModRuleActionType.send_alert_message.value:
            channel_id = int(data["metadata"]["channel_id"])
            return cls(channel_id=channel_id)
        return cls(custom_message=data.get("metadata", {}).get("custom_message"))

    def to_dict(self) -> Dict[str, Any]:
        ret = {"type": self.type.value, "metadata": {}}
        if (
            self.type is AutoModRuleActionType.block_message
            and self.custom_message is not None
        ):
            ret["metadata"] = {"custom_message": self.custom_message}
        elif self.type is AutoModRuleActionType.timeout:
            ret["metadata"] = {"duration_seconds": int(self.duration.total_seconds())}  # type: ignore # duration cannot be None here
        elif self.type is AutoModRuleActionType.send_alert_message:
            ret["metadata"] = {"channel_id": str(self.channel_id)}
        return ret


class AutoModTrigger:
    r"""Represents a trigger for an auto moderation rule.

    The following table illustrates relevant attributes for each :class:`AutoModRuleTriggerType`:

    +-----------------------------------------------+------------------------------------------------+
    |                    Type                       |                   Attributes                   |
    +===============================================+================================================+
    | :attr:`AutoModRuleTriggerType.keyword`        | :attr:`keyword_filter`, :attr:`regex_patterns`,|
    |                                               | :attr:`allow_list`                             |
    +-----------------------------------------------+------------------------------------------------+
    | :attr:`AutoModRuleTriggerType.spam`           |                                                |
    +-----------------------------------------------+------------------------------------------------+
    | :attr:`AutoModRuleTriggerType.keyword_preset` | :attr:`presets`\, :attr:`allow_list`           |
    +-----------------------------------------------+------------------------------------------------+
    | :attr:`AutoModRuleTriggerType.mention_spam`   | :attr:`mention_limit`                          |
    +-----------------------------------------------+------------------------------------------------+

    .. versionadded:: 2.0

    Attributes
    -----------
    type: :class:`AutoModRuleTriggerType`
        The type of trigger.
    keyword_filter: List[:class:`str`]
        The list of strings that will trigger the keyword filter. Maximum of 1000.
        Keywords can only be up to 60 characters in length.

        This could be combined with :attr:`regex_patterns`.
    regex_patterns: List[:class:`str`]
        The regex pattern that will trigger the filter. The syntax is based off of
        `Rust's regex syntax <https://docs.rs/regex/latest/regex/#syntax>`_.
        Maximum of 10. Regex strings can only be up to 260 characters in length.

        This could be combined with :attr:`keyword_filter` and/or :attr:`allow_list`

        .. versionadded:: 2.0
    presets: :class:`AutoModPresets`
        The presets used with the preset keyword filter.
    allow_list: List[:class:`str`]
        The list of words that are exempt from the commonly flagged words. Maximum of 100.
        Keywords can only be up to 60 characters in length.
    mention_limit: :class:`int`
        The total number of user and role mentions a message can contain.
        Has a maximum of 50.
    """

    __slots__ = (
        "type",
        "keyword_filter",
        "presets",
        "allow_list",
        "mention_limit",
        "regex_patterns",
    )

    def __init__(
        self,
        *,
        type: Optional[AutoModRuleTriggerType] = None,
        keyword_filter: Optional[List[str]] = None,
        presets: Optional[AutoModPresets] = None,
        allow_list: Optional[List[str]] = None,
        mention_limit: Optional[int] = None,
        regex_patterns: Optional[List[str]] = None,
    ) -> None:
        if (
            type is None
            and sum(
                arg is not None
                for arg in (keyword_filter or regex_patterns, presets, mention_limit)
            )
            > 1
        ):
            raise ValueError(
                "Please pass only one of keyword_filter, regex_patterns, presets, or mention_limit."
            )

        if type is not None:
            self.type = type
        elif keyword_filter is not None or regex_patterns is not None:
            self.type = AutoModRuleTriggerType.keyword
        elif presets is not None:
            self.type = AutoModRuleTriggerType.keyword_preset
        elif mention_limit is not None:
            self.type = AutoModRuleTriggerType.mention_spam
        else:
            raise ValueError(
                "Please pass the trigger type explicitly if not using keyword_filter, presets, or mention_limit."
            )

        self.keyword_filter: List[str] = (
            keyword_filter if keyword_filter is not None else []
        )
        self.presets: AutoModPresets = (
            presets if presets is not None else AutoModPresets()
        )
        self.allow_list: List[str] = allow_list if allow_list is not None else []
        self.mention_limit: int = mention_limit if mention_limit is not None else 0
        self.regex_patterns: List[str] = (
            regex_patterns if regex_patterns is not None else []
        )

    def __repr__(self) -> str:
        data = self.to_metadata_dict()
        if data:
            joined = " ".join(f"{k}={v!r}" for k, v in data.items())
            return f"<AutoModTrigger type={self.type} {joined}>"

        return f"<AutoModTrigger type={self.type}>"

    @classmethod
    def from_data(
        cls, type: int, data: Optional[AutoModerationTriggerMetadataPayload]
    ) -> Self:
        type_ = try_enum(AutoModRuleTriggerType, type)
        if data is None:
            return cls(type=type_)
        elif type_ is AutoModRuleTriggerType.keyword:
            return cls(
                type=type_,
                keyword_filter=data.get("keyword_filter"),
                regex_patterns=data.get("regex_patterns"),
                allow_list=data.get("allow_list"),
            )
        elif type_ is AutoModRuleTriggerType.keyword_preset:
            return cls(
                type=type_,
                presets=AutoModPresets._from_value(data.get("presets", [])),
                allow_list=data.get("allow_list"),
            )
        elif type_ is AutoModRuleTriggerType.mention_spam:
            return cls(type=type_, mention_limit=data.get("mention_total_limit"))
        else:
            return cls(type=type_)

    def to_metadata_dict(self) -> Optional[Dict[str, Any]]:
        if self.type is AutoModRuleTriggerType.keyword:
            return {
                "keyword_filter": self.keyword_filter,
                "regex_patterns": self.regex_patterns,
                "allow_list": self.allow_list,
            }
        elif self.type is AutoModRuleTriggerType.keyword_preset:
            return {"presets": self.presets.to_array(), "allow_list": self.allow_list}
        elif self.type is AutoModRuleTriggerType.mention_spam:
            return {"mention_total_limit": self.mention_limit}


class AutoModRule:
    """Represents an auto moderation rule.

    .. versionadded:: 2.0

    Attributes
    -----------
    id: :class:`int`
        The ID of the rule.
    guild: :class:`Guild`
        The guild the rule is for.
    name: :class:`str`
        The name of the rule.
    creator_id: :class:`int`
        The ID of the user that created the rule.
    trigger: :class:`AutoModTrigger`
        The rule's trigger.
    enabled: :class:`bool`
        Whether the rule is enabled.
    exempt_role_ids: Set[:class:`int`]
        The IDs of the roles that are exempt from the rule.
    exempt_channel_ids: Set[:class:`int`]
        The IDs of the channels that are exempt from the rule.
    """

    __slots__ = (
        "_state",
        "_cs_exempt_roles",
        "_cs_exempt_channels",
        "_cs_actions",
        "id",
        "guild",
        "name",
        "creator_id",
        "event_type",
        "trigger",
        "enabled",
        "exempt_role_ids",
        "exempt_channel_ids",
        "_actions",
    )

    def __init__(
        self, *, data: AutoModerationRulePayload, guild: Guild, state: ConnectionState
    ) -> None:
        self._state: ConnectionState = state
        self.guild: Guild = guild
        self.id: int = int(data["id"])
        self.name: str = data["name"]
        self.creator_id = int(data["creator_id"])
        self.event_type: AutoModRuleEventType = try_enum(
            AutoModRuleEventType, data["event_type"]
        )
        self.trigger: AutoModTrigger = AutoModTrigger.from_data(
            data["trigger_type"], data=data.get("trigger_metadata")
        )
        self.enabled: bool = data["enabled"]
        self.exempt_role_ids: Set[int] = {
            int(role_id) for role_id in data["exempt_roles"]
        }
        self.exempt_channel_ids: Set[int] = {
            int(channel_id) for channel_id in data["exempt_channels"]
        }
        self._actions: List[AutoModerationActionPayload] = data["actions"]

    def __repr__(self) -> str:
        return f"<AutoModRule id={self.id} name={self.name!r} guild={self.guild!r}>"

    def to_dict(self) -> AutoModerationRulePayload:
        ret: AutoModerationRulePayload = {
            "id": str(self.id),
            "guild_id": str(self.guild.id),
            "name": self.name,
            "creator_id": str(self.creator_id),
            "event_type": self.event_type.value,
            "trigger_type": self.trigger.type.value,
            "trigger_metadata": self.trigger.to_metadata_dict(),
            "actions": [action.to_dict() for action in self.actions],
            "enabled": self.enabled,
            "exempt_roles": [str(role_id) for role_id in self.exempt_role_ids],
            "exempt_channels": [
                str(channel_id) for channel_id in self.exempt_channel_ids
            ],
        }  # type: ignore # trigger types break the flow here.

        return ret

    @property
    def creator(self) -> Optional[Member]:
        """Optional[:class:`Member`]: The member that created this rule."""
        return self.guild.get_member(self.creator_id)

    @cached_slot_property("_cs_exempt_roles")
    def exempt_roles(self) -> List[Role]:
        """List[:class:`Role`]: The roles that are exempt from this rule."""
        result = []
        get_role = self.guild.get_role
        for role_id in self.exempt_role_ids:
            role = get_role(role_id)
            if role is not None:
                result.append(role)

        return utils._unique(result)

    @cached_slot_property("_cs_exempt_channels")
    def exempt_channels(self) -> List[Union[GuildChannel, Thread]]:
        """List[Union[:class:`abc.GuildChannel`, :class:`Thread`]]: The channels that are exempt from this rule."""
        it = filter(None, map(self.guild._resolve_channel, self.exempt_channel_ids))
        return utils._unique(it)

    @cached_slot_property("_cs_actions")
    def actions(self) -> List[AutoModRuleAction]:
        """List[:class:`AutoModRuleAction`]: The actions that are taken when this rule is triggered."""
        return [AutoModRuleAction.from_data(action) for action in self._actions]

    def is_exempt(self, obj: Snowflake, /) -> bool:
        """Check if an object is exempt from the automod rule.

        Parameters
        -----------
        obj: :class:`abc.Snowflake`
            The role, channel, or thread to check.

        Returns
        --------
        :class:`bool`
            Whether the object is exempt from the automod rule.
        """
        return obj.id in self.exempt_channel_ids or obj.id in self.exempt_role_ids

    async def edit(
        self,
        *,
        name: str = MISSING,
        event_type: AutoModRuleEventType = MISSING,
        actions: List[AutoModRuleAction] = MISSING,
        trigger: AutoModTrigger = MISSING,
        enabled: bool = MISSING,
        exempt_roles: Sequence[Snowflake] = MISSING,
        exempt_channels: Sequence[Snowflake] = MISSING,
        reason: str = MISSING,
    ) -> Self:
        """|coro|

        Edits this auto moderation rule.

        You must have :attr:`Permissions.manage_guild` to edit rules.

        Parameters
        -----------
        name: :class:`str`
            The new name to change to.
        event_type: :class:`AutoModRuleEventType`
            The new event type to change to.
        actions: List[:class:`AutoModRuleAction`]
            The new rule actions to update.
        trigger: :class:`AutoModTrigger`
            The new trigger to update.
            You can only change the trigger metadata, not the type.
        enabled: :class:`bool`
            Whether the rule should be enabled or not.
        exempt_roles: Sequence[:class:`abc.Snowflake`]
            The new roles to exempt from the rule.
        exempt_channels: Sequence[:class:`abc.Snowflake`]
            The new channels to exempt from the rule.
        reason: :class:`str`
            The reason for updating this rule. Shows up on the audit log.

        Raises
        -------
        Forbidden
            You do not have permission to edit this rule.
        HTTPException
            Editing the rule failed.

        Returns
        --------
        :class:`AutoModRule`
            The updated auto moderation rule.
        """
        payload = {}
        if actions is not MISSING:
            payload["actions"] = [action.to_dict() for action in actions]

        if name is not MISSING:
            payload["name"] = name

        if event_type is not MISSING:
            payload["event_type"] = event_type

        if trigger is not MISSING:
            trigger_metadata = trigger.to_metadata_dict()
            if trigger_metadata is not None:
                payload["trigger_metadata"] = trigger_metadata

        if enabled is not MISSING:
            payload["enabled"] = enabled

        if exempt_roles is not MISSING:
            payload["exempt_roles"] = [x.id for x in exempt_roles]

        if exempt_channels is not MISSING:
            payload["exempt_channels"] = [x.id for x in exempt_channels]

        data = await self._state.http.edit_auto_moderation_rule(
            self.guild.id,
            self.id,
            reason=reason,
            **payload,
        )

        return AutoModRule(data=data, guild=self.guild, state=self._state)

    async def delete(self, *, reason: str = MISSING) -> None:
        """|coro|

        Deletes the auto moderation rule.

        You must have :attr:`Permissions.manage_guild` to delete rules.

        Parameters
        -----------
        reason: :class:`str`
            The reason for deleting this rule. Shows up on the audit log.

        Raises
        -------
        Forbidden
            You do not have permissions to delete the rule.
        HTTPException
            Deleting the rule failed.
        """
        await self._state.http.delete_auto_moderation_rule(
            self.guild.id, self.id, reason=reason
        )


class AutoModAction:
    """Represents an action that was taken as the result of a moderation rule.

    .. versionadded:: 2.0

    Attributes
    -----------
    action: :class:`AutoModRuleAction`
        The action that was taken.
    message_id: Optional[:class:`int`]
        The message ID that triggered the action. This is only available if the
        action is done on an edited message.
    rule_id: :class:`int`
        The ID of the rule that was triggered.
    rule_trigger_type: :class:`AutoModRuleTriggerType`
        The trigger type of the rule that was triggered.
    guild_id: :class:`int`
        The ID of the guild where the rule was triggered.
    user_id: :class:`int`
        The ID of the user that triggered the rule.
    channel_id: :class:`int`
        The ID of the channel where the rule was triggered.
    alert_system_message_id: Optional[:class:`int`]
        The ID of the system message that was sent to the predefined alert channel.
    content: :class:`str`
        The content of the message that triggered the rule.
    matched_keyword: Optional[:class:`str`]
        The matched keyword from the triggering message.
    matched_content: Optional[:class:`str`]
        The matched content from the triggering message.
    """

    __slots__ = (
        "_state",
        "action",
        "rule_id",
        "rule_trigger_type",
        "guild_id",
        "user_id",
        "channel_id",
        "message_id",
        "alert_system_message_id",
        "content",
        "matched_keyword",
        "matched_content",
    )

    def __init__(
        self, *, data: AutoModerationActionExecutionPayload, state: ConnectionState
    ) -> None:
        self._state: ConnectionState = state
        self.message_id: Optional[int] = utils._get_as_snowflake(data, "message_id")
        self.action: AutoModRuleAction = AutoModRuleAction.from_data(data["action"])
        self.rule_id: int = int(data["rule_id"])
        self.rule_trigger_type: AutoModRuleTriggerType = try_enum(
            AutoModRuleTriggerType, data["rule_trigger_type"]
        )
        self.guild_id: int = int(data["guild_id"])
        self.channel_id: Optional[int] = utils._get_as_snowflake(data, "channel_id")
        self.user_id: int = int(data["user_id"])
        self.alert_system_message_id: Optional[int] = utils._get_as_snowflake(
            data, "alert_system_message_id"
        )
        self.content: str = data.get("content", "")
        self.matched_keyword: Optional[str] = data["matched_keyword"]
        self.matched_content: Optional[str] = data.get("matched_content")

    def __repr__(self) -> str:
        return f"<AutoModRuleExecution rule_id={self.rule_id} action={self.action!r}>"

    @property
    def guild(self) -> Guild:
        """:class:`Guild`: The guild this action was taken in."""
        return self._state._get_or_create_unavailable_guild(self.guild_id)

    @property
    def channel(self) -> Optional[Union[GuildChannel, Thread]]:
        """Optional[Union[:class:`abc.GuildChannel`, :class:`Thread`]]: The channel this action was taken in."""
        if self.channel_id:
            return self.guild.get_channel_or_thread(self.channel_id)
        return None

    @property
    def member(self) -> Optional[Member]:
        """Optional[:class:`Member`]: The member this action was taken against /who triggered this rule."""
        return self.guild.get_member(self.user_id)

    async def fetch_rule(self) -> AutoModRule:
        """|coro|

        Fetch the rule whose action was taken.

        You must have :attr:`Permissions.manage_guild` to do this.

        Raises
        -------
        Forbidden
            You do not have permissions to view the rule.
        HTTPException
            Fetching the rule failed.

        Returns
        --------
        :class:`AutoModRule`
            The rule that was executed.
        """

        data = await self._state.http.get_auto_moderation_rule(
            self.guild.id, self.rule_id
        )
        return AutoModRule(data=data, guild=self.guild, state=self._state)



from discord_self._vendor.discord.guild_premium import (
    PremiumGuildSubscription,
    PremiumGuildSubscriptionCooldown,
    PremiumGuildSubscriptionSlot,
)

__all__ = [
    "PremiumGuildSubscription",
    "PremiumGuildSubscriptionCooldown",
    "PremiumGuildSubscriptionSlot",
]



from discord_self._vendor.discord.types.activity import Activity
from discord_self._vendor.discord.types.snowflake import Snowflake
from discord_self._vendor.discord.types.user import User
from discord_self._vendor.discord.types.widget import (
    EditWidgetSettings,
    Widget,
    WidgetChannel,
    WidgetMember,
    WidgetSettings,
)

__all__ = [
    "Activity",
    "EditWidgetSettings",
    "Snowflake",
    "User",
    "Widget",
    "WidgetChannel",
    "WidgetMember",
    "WidgetSettings",
]



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from typing import List, Literal, Optional, TypedDict, Union

from typing_extensions import NotRequired

from .snowflake import Snowflake

AutoModerationRuleTriggerType = Literal[1, 2, 3, 4]
AutoModerationActionTriggerType = Literal[1, 2, 3]
AutoModerationRuleEventType = Literal[1]
AutoModerationTriggerPresets = Literal[1, 2, 3]


class Empty(TypedDict): ...


class _AutoModerationActionMetadataAlert(TypedDict):
    channel_id: Snowflake


class _AutoModerationActionMetadataTimeout(TypedDict):
    duration_seconds: int


class _AutoModerationActionMetadataCustomMessage(TypedDict):
    custom_message: str


class _AutoModerationActionBlockMessage(TypedDict):
    type: Literal[1]
    metadata: NotRequired[_AutoModerationActionMetadataCustomMessage]


class _AutoModerationActionAlert(TypedDict):
    type: Literal[2]
    metadata: _AutoModerationActionMetadataAlert


class _AutoModerationActionTimeout(TypedDict):
    type: Literal[3]
    metadata: _AutoModerationActionMetadataTimeout


AutoModerationAction = Union[
    _AutoModerationActionBlockMessage,
    _AutoModerationActionAlert,
    _AutoModerationActionTimeout,
]


class _AutoModerationTriggerMetadataKeyword(TypedDict):
    keyword_filter: List[str]
    regex_patterns: NotRequired[List[str]]


class _AutoModerationTriggerMetadataKeywordPreset(TypedDict):
    presets: List[AutoModerationTriggerPresets]
    allow_list: List[str]


class _AutoModerationTriggerMetadataMentionLimit(TypedDict):
    mention_total_limit: int


AutoModerationTriggerMetadata = Union[
    _AutoModerationTriggerMetadataKeyword,
    _AutoModerationTriggerMetadataKeywordPreset,
    _AutoModerationTriggerMetadataMentionLimit,
    Empty,
]


class _BaseAutoModerationRule(TypedDict):
    id: Snowflake
    guild_id: Snowflake
    name: str
    creator_id: Snowflake
    event_type: AutoModerationRuleEventType
    actions: List[AutoModerationAction]
    enabled: bool
    exempt_roles: List[Snowflake]
    exempt_channels: List[Snowflake]


class _AutoModerationRuleKeyword(_BaseAutoModerationRule):
    trigger_type: Literal[1]
    trigger_metadata: _AutoModerationTriggerMetadataKeyword


class _AutoModerationRuleKeywordPreset(_BaseAutoModerationRule):
    trigger_type: Literal[4]
    trigger_metadata: _AutoModerationTriggerMetadataKeywordPreset


class _AutoModerationRuleOther(_BaseAutoModerationRule):
    trigger_type: Literal[2, 3]


AutoModerationRule = Union[
    _AutoModerationRuleKeyword,
    _AutoModerationRuleKeywordPreset,
    _AutoModerationRuleOther,
]


class AutoModerationActionExecution(TypedDict):
    guild_id: Snowflake
    action: AutoModerationAction
    rule_id: Snowflake
    rule_trigger_type: AutoModerationRuleTriggerType
    user_id: Snowflake
    channel_id: NotRequired[Snowflake]
    message_id: NotRequired[Snowflake]
    alert_system_message_id: NotRequired[Snowflake]
    content: str
    matched_keyword: Optional[str]
    matched_content: Optional[str]



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional, Union

from .asset import Asset
from .enums import (
    ChannelType,
    InviteTarget,
    InviteType,
    NSFWLevel,
    VerificationLevel,
    try_enum,
)
from .mixins import Hashable
from .object import Object
from .scheduled_event import ScheduledEvent
from .utils import MISSING, _get_as_snowflake, parse_time, snowflake_time
from .welcome_screen import WelcomeScreen

__all__ = (
    "PartialInviteChannel",
    "PartialInviteGuild",
    "Invite",
)

if TYPE_CHECKING:
    from typing_extensions import Self

    from .abc import GuildChannel, Snowflake
    from .application import PartialApplication
    from .channel import DMChannel, GroupChannel
    from .guild import Guild
    from .message import Message
    from .state import ConnectionState
    from .types.channel import (
        PartialChannel as InviteChannelPayload,
    )
    from .types.invite import (
        GatewayInvite as GatewayInvitePayload,
    )
    from .types.invite import (
        Invite as InvitePayload,
    )
    from .types.invite import (
        InviteGuild as InviteGuildPayload,
    )
    from .user import User

    InviteGuildType = Union[Guild, "PartialInviteGuild", Object]
    InviteChannelType = Union[
        GuildChannel, "PartialInviteChannel", Object, DMChannel, GroupChannel
    ]

    import datetime


class PartialInviteChannel:
    """Represents a "partial" invite channel.

    This model will be given when the user is not part of the
    guild or group channel the :class:`Invite` resolves to.

    .. container:: operations

        .. describe:: x == y

            Checks if two partial channels are the same.

        .. describe:: x != y

            Checks if two partial channels are not the same.

        .. describe:: hash(x)

            Return the partial channel's hash.

        .. describe:: str(x)

            Returns the partial channel's name.

    Attributes
    -----------
    name: Optional[:class:`str`]
        The partial channel's name.
    id: :class:`int`
        The partial channel's ID.
    type: :class:`ChannelType`
        The partial channel's type.
    recipients: Optional[List[:class:`str`]]
        The partial channel's recipient names. This is only applicable to group DMs.

        .. versionadded:: 2.0
    """

    __slots__ = ("_state", "id", "name", "type", "recipients", "_icon")

    def __new__(cls, data: Optional[InviteChannelPayload], *args, **kwargs):
        if data is None:
            return
        return super().__new__(cls)

    def __init__(self, data: Optional[InviteChannelPayload], state: ConnectionState):
        if data is None:
            return
        self._state = state
        self.id: int = int(data["id"])
        self.name: Optional[str] = data.get("name")
        self.type: ChannelType = try_enum(ChannelType, data["type"])
        self.recipients: Optional[List[str]] = (
            [user["username"] for user in data.get("recipients", [])]
            if self.type in (ChannelType.private, ChannelType.group)
            else None
        )
        self._icon: Optional[str] = data.get("icon")

    def __str__(self) -> str:
        if self.name:
            return self.name

        recipients = self.recipients or []
        if self.type == ChannelType.group:
            return ", ".join(recipients) if recipients else "Unnamed"
        return f"Direct Message with {recipients[0] if recipients else 'Unknown User'}"

    def __repr__(self) -> str:
        return (
            f"<PartialInviteChannel id={self.id} name={self.name} type={self.type!r}>"
        )

    @property
    def mention(self) -> str:
        """:class:`str`: The string that allows you to mention the channel."""
        return f"<#{self.id}>"

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns the channel's creation time in UTC."""
        return snowflake_time(self.id)

    @property
    def icon(self) -> Optional[Asset]:
        """Optional[:class:`Asset`]: Returns the channel's icon asset if available.

        Only applicable to channels of type :attr:`ChannelType.group`.

        .. versionadded:: 2.0
        """
        if self._icon is None:
            return None
        return Asset._from_icon(self._state, self.id, self._icon, path="channel")


class PartialInviteGuild:
    """Represents a "partial" invite guild.

    This model will be given when the user is not part of the
    guild the :class:`Invite` resolves to.

    .. container:: operations

        .. describe:: x == y

            Checks if two partial guilds are the same.

        .. describe:: x != y

            Checks if two partial guilds are not the same.

        .. describe:: hash(x)

            Return the partial guild's hash.

        .. describe:: str(x)

            Returns the partial guild's name.

    Attributes
    -----------
    name: :class:`str`
        The partial guild's name.
    id: :class:`int`
        The partial guild's ID.
    verification_level: :class:`VerificationLevel`
        The partial guild's verification level.
    features: List[:class:`str`]
        A list of features the guild has. See :attr:`Guild.features` for more information.
    description: Optional[:class:`str`]
        The partial guild's description.
    nsfw_level: :class:`NSFWLevel`
        The partial guild's NSFW level.

        .. versionadded:: 2.0
    vanity_url_code: Optional[:class:`str`]
        The partial guild's vanity URL code, if available.

        .. versionadded:: 2.0
    premium_subscription_count: :class:`int`
        The number of "boosts" the partial guild currently has.

        .. versionadded:: 2.0
    """

    __slots__ = (
        "_state",
        "_icon",
        "_banner",
        "_splash",
        "features",
        "id",
        "name",
        "verification_level",
        "description",
        "vanity_url_code",
        "nsfw_level",
        "premium_subscription_count",
    )

    def __init__(self, state: ConnectionState, data: InviteGuildPayload, id: int):
        self._state: ConnectionState = state
        self.id: int = id
        self.name: str = data["name"]
        self.features: List[str] = data.get("features", [])
        self._icon: Optional[str] = data.get("icon")
        self._banner: Optional[str] = data.get("banner")
        self._splash: Optional[str] = data.get("splash")
        self.verification_level: VerificationLevel = try_enum(
            VerificationLevel, data.get("verification_level")
        )
        self.description: Optional[str] = data.get("description")
        self.vanity_url_code: Optional[str] = data.get("vanity_url_code")
        self.nsfw_level: NSFWLevel = try_enum(NSFWLevel, data.get("nsfw_level", 0))
        self.premium_subscription_count: int = (
            data.get("premium_subscription_count") or 0
        )

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} id={self.id} name={self.name!r} features={self.features} "
            f"description={self.description!r}>"
        )

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns the guild's creation time in UTC."""
        return snowflake_time(self.id)

    @property
    def vanity_url(self) -> Optional[str]:
        """Optional[:class:`str`]: The Discord vanity invite URL for this partial guild, if available.

        .. versionadded:: 2.0
        """
        if self.vanity_url_code is None:
            return None
        return f"{Invite.BASE}/{self.vanity_url_code}"

    @property
    def icon(self) -> Optional[Asset]:
        """Optional[:class:`Asset`]: Returns the guild's icon asset, if available."""
        if self._icon is None:
            return None
        return Asset._from_guild_icon(self._state, self.id, self._icon)

    @property
    def banner(self) -> Optional[Asset]:
        """Optional[:class:`Asset`]: Returns the guild's banner asset, if available."""
        if self._banner is None:
            return None
        return Asset._from_guild_image(
            self._state, self.id, self._banner, path="banners"
        )

    @property
    def splash(self) -> Optional[Asset]:
        """Optional[:class:`Asset`]: Returns the guild's invite splash asset, if available."""
        if self._splash is None:
            return None
        return Asset._from_guild_image(
            self._state, self.id, self._splash, path="splashes"
        )


class Invite(Hashable):
    r"""Represents a Discord :class:`Guild` or :class:`abc.GuildChannel` invite.

    Depending on the way this object was created, some of the attributes can
    have a value of ``None``.

    .. container:: operations

        .. describe:: x == y

            Checks if two invites are equal.

        .. describe:: x != y

            Checks if two invites are not equal.

        .. describe:: hash(x)

            Returns the invite hash.

        .. describe:: str(x)

            Returns the invite URL.

    The following table illustrates what methods will obtain the attributes:

    +------------------------------------+--------------------------------------------------------------+
    |             Attribute              |                          Method                              |
    +====================================+==============================================================+
    | :attr:`max_age`                    | :meth:`abc.GuildChannel.invites`\, :meth:`Guild.invites`     |
    +------------------------------------+--------------------------------------------------------------+
    | :attr:`max_uses`                   | :meth:`abc.GuildChannel.invites`\, :meth:`Guild.invites`     |
    +------------------------------------+--------------------------------------------------------------+
    | :attr:`created_at`                 | :meth:`abc.GuildChannel.invites`\, :meth:`Guild.invites`     |
    +------------------------------------+--------------------------------------------------------------+
    | :attr:`temporary`                  | :meth:`abc.GuildChannel.invites`\, :meth:`Guild.invites`     |
    +------------------------------------+--------------------------------------------------------------+
    | :attr:`uses`                       | :meth:`abc.GuildChannel.invites`\, :meth:`Guild.invites`     |
    +------------------------------------+--------------------------------------------------------------+
    | :attr:`approximate_member_count`   | :meth:`Client.fetch_invite` with ``with_counts`` enabled     |
    +------------------------------------+--------------------------------------------------------------+
    | :attr:`approximate_presence_count` | :meth:`Client.fetch_invite` with ``with_counts`` enabled     |
    +------------------------------------+--------------------------------------------------------------+

    If it's not in the table above then it is available by all methods.

    Attributes
    -----------
    max_age: Optional[:class:`int`]
        How long before the invite expires in seconds.
        A value of ``0`` indicates that it doesn't expire.
    code: :class:`str`
        The URL fragment used for the invite.
    type: :class:`InviteType`
        The type of invite.

        .. versionadded:: 2.0
    guild: Optional[Union[:class:`Guild`, :class:`Object`, :class:`PartialInviteGuild`]]
        The guild the invite is for. Can be ``None`` if not a guild invite.
    revoked: Optional[:class:`bool`]
        Indicates if the invite has been revoked.
    created_at: Optional[:class:`datetime.datetime`]
        An aware UTC datetime object denoting the time the invite was created.
    temporary: Optional[:class:`bool`]
        Indicates that the invite grants temporary membership.
        If ``True``, members who joined via this invite will be kicked upon disconnect.
    uses: Optional[:class:`int`]
        How many times the invite has been used.
    max_uses: Optional[:class:`int`]
        How many times the invite can be used.
        A value of ``0`` indicates that it has unlimited uses.
    inviter: Optional[:class:`User`]
        The user who created the invite.
    approximate_member_count: Optional[:class:`int`]
        The approximate number of members in the guild.
    approximate_presence_count: Optional[:class:`int`]
        The approximate number of members currently active in the guild.
        This includes idle, dnd, online, and invisible members. Offline members are excluded.
    expires_at: Optional[:class:`datetime.datetime`]
        The expiration date of the invite. If the value is ``None`` (unless received through
        :meth:`Client.fetch_invite` with ``with_expiration`` disabled), the invite will never expire.

        .. versionadded:: 2.0
    channel: Optional[Union[:class:`abc.GuildChannel`, :class:`GroupChannel`, :class:`Object`, :class:`PartialInviteChannel`]]
        The channel the invite is for. Can be ``None`` if not a guild invite.
    target_type: :class:`InviteTarget`
        The type of target for the voice channel invite.

        .. versionadded:: 2.0
    target_user: Optional[:class:`User`]
        The user whose stream to display for this invite, if any.

        .. versionadded:: 2.0
    target_application: Optional[:class:`PartialApplication`]
        The embedded application the invite targets, if any.

        .. versionadded:: 2.0
    scheduled_event: Optional[:class:`ScheduledEvent`]
        The scheduled event associated with this invite, if any.

        .. versionadded:: 2.0
    scheduled_event_id: Optional[:class:`int`]
        The ID of the scheduled event associated with this invite, if any.

        .. versionadded:: 2.0
    welcome_screen: Optional[:class:`WelcomeScreen`]
        The guild's welcome screen, if available.

        .. versionadded:: 2.0
    new_member: :class:`bool`
        Whether the user was not previously a member of the guild.

        .. versionadded:: 2.0

        .. note::
            This is only possibly ``True`` in accepted invite objects
            (i.e. the objects received from :meth:`accept` and :meth:`use`).
    show_verification_form: :class:`bool`
        Whether the user should be shown the guild's member verification form.

        .. versionadded:: 2.0

        .. note::
            This is only possibly ``True`` in accepted invite objects
            (i.e. the objects received from :meth:`accept` and :meth:`use`).
    """

    __slots__ = (
        "max_age",
        "code",
        "guild",
        "revoked",
        "created_at",
        "uses",
        "temporary",
        "max_uses",
        "inviter",
        "channel",
        "target_user",
        "target_type",
        "_state",
        "approximate_member_count",
        "approximate_presence_count",
        "target_application",
        "expires_at",
        "scheduled_event",
        "scheduled_event_id",
        "_message",
        "welcome_screen",
        "type",
        "new_member",
        "show_verification_form",
    )

    BASE = "https://discord.gg"

    def __init__(
        self,
        *,
        state: ConnectionState,
        data: InvitePayload,
        guild: Optional[Union[PartialInviteGuild, Guild]] = None,
        channel: Optional[
            Union[PartialInviteChannel, GuildChannel, GroupChannel]
        ] = None,
        welcome_screen: Optional[WelcomeScreen] = None,
    ):
        self._state: ConnectionState = state
        self.type: InviteType = try_enum(InviteType, data.get("type", 0))
        self.max_age: Optional[int] = data.get("max_age")
        self.code: str = data["code"]
        self.guild: Optional[InviteGuildType] = self._resolve_guild(
            data.get("guild"), guild
        )
        self.revoked: Optional[bool] = data.get("revoked")
        self.created_at: Optional[datetime.datetime] = parse_time(
            data.get("created_at")
        )
        self.temporary: Optional[bool] = data.get("temporary")
        self.uses: Optional[int] = data.get("uses")
        self.max_uses: Optional[int] = data.get("max_uses")
        self.approximate_presence_count: Optional[int] = data.get(
            "approximate_presence_count"
        )
        self.approximate_member_count: Optional[int] = data.get(
            "approximate_member_count"
        )
        self._message: Optional[Message] = data.get("message")

        # We inject some missing data here since we can assume it
        if self.type in (InviteType.group_dm, InviteType.friend):
            self.temporary = False
            if self.max_uses is None and self.type is InviteType.group_dm:
                self.max_uses = 0

        expires_at = data.get("expires_at", None)
        self.expires_at: Optional[datetime.datetime] = (
            parse_time(expires_at) if expires_at else None
        )

        inviter_data = data.get("inviter")
        self.inviter: Optional[User] = (
            None if inviter_data is None else self._state.create_user(inviter_data)
        )

        self.channel: Optional[InviteChannelType] = self._resolve_channel(
            data.get("channel"), channel
        )

        target_user_data = data.get("target_user")
        self.target_user: Optional[User] = (
            None
            if target_user_data is None
            else self._state.create_user(target_user_data)
        )

        self.target_type: InviteTarget = try_enum(
            InviteTarget, data.get("target_type", 0)
        )

        application = data.get("target_application")
        if application is not None:
            from .application import PartialApplication

            application = PartialApplication(data=application, state=state)
        self.target_application: Optional[PartialApplication] = application

        self.welcome_screen = welcome_screen

        scheduled_event = data.get("guild_scheduled_event")
        self.scheduled_event: Optional[ScheduledEvent] = (
            ScheduledEvent(
                state=self._state,
                data=scheduled_event,
            )
            if scheduled_event
            else None
        )
        self.scheduled_event_id: Optional[int] = (
            self.scheduled_event.id if self.scheduled_event else None
        )

        # Only present on accepted invites
        self.new_member: bool = data.get("new_member", False)
        self.show_verification_form: bool = data.get("show_verification_form", False)

    @classmethod
    def from_incomplete(
        cls,
        *,
        state: ConnectionState,
        data: InvitePayload,
        message: Optional[Message] = None,
    ) -> Self:
        guild: Optional[Union[Guild, PartialInviteGuild]]
        try:
            guild_data = data["guild"]
        except KeyError:
            # If we're here, then this is a group DM
            guild = None
            welcome_screen = None
        else:
            guild_id = int(guild_data["id"])
            guild = state._get_guild(guild_id)
            if guild is None:
                guild = PartialInviteGuild(state, guild_data, guild_id)

            welcome_screen = guild_data.get("welcome_screen")
            if welcome_screen is not None:
                welcome_screen = WelcomeScreen(data=welcome_screen, guild=guild)

        channel_data = data.get("channel")
        if channel_data and channel_data.get("type") == ChannelType.private.value:
            channel_data["recipients"] = [data["inviter"]] if "inviter" in data else []
        channel = PartialInviteChannel(channel_data, state)
        channel = state.get_channel(getattr(channel, "id", None)) or channel

        if message is not None:
            data["message"] = message  # type: ignore # Not a real field

        return cls(
            state=state,
            data=data,
            guild=guild,
            channel=channel,
            welcome_screen=welcome_screen,
        )  # type: ignore

    @classmethod
    def from_gateway(
        cls, *, state: ConnectionState, data: GatewayInvitePayload
    ) -> Self:
        guild_id: Optional[int] = _get_as_snowflake(data, "guild_id")
        guild: Optional[Union[Guild, Object]] = state._get_guild(guild_id)
        channel_id = _get_as_snowflake(data, "channel_id")
        if guild is not None:
            channel = (
                (guild.get_channel(channel_id) or Object(id=channel_id))
                if channel_id is not None
                else None
            )
        else:
            guild = (
                state._get_or_create_unavailable_guild(guild_id)
                if guild_id is not None
                else None
            )
            channel = Object(id=channel_id) if channel_id is not None else None

        return cls(state=state, data=data, guild=guild, channel=channel)  # type: ignore

    def _resolve_guild(
        self,
        data: Optional[InviteGuildPayload],
        guild: Optional[Union[Guild, PartialInviteGuild]] = None,
    ) -> Optional[InviteGuildType]:
        if guild is not None:
            return guild

        if data is None:
            return None

        guild_id = int(data["id"])
        return PartialInviteGuild(self._state, data, guild_id)

    def _resolve_channel(
        self,
        data: Optional[InviteChannelPayload],
        channel: Optional[
            Union[PartialInviteChannel, GuildChannel, GroupChannel]
        ] = None,
    ) -> Optional[InviteChannelType]:
        if channel is not None:
            return channel

        if data is None:
            return None

        return PartialInviteChannel(data, self._state)

    def __str__(self) -> str:
        return self.url

    def __repr__(self) -> str:
        return (
            f"<Invite code={self.code!r} type={self.type!r} "
            f"guild={self.guild!r} "
            f"members={self.approximate_member_count}>"
        )

    def __hash__(self) -> int:
        return hash(self.code)

    @property
    def id(self) -> str:
        """:class:`str`: Returns the proper code portion of the invite."""
        return self.code

    @property
    def url(self) -> str:
        """:class:`str`: A property that retrieves the invite URL."""
        url = self.BASE + "/" + self.code
        if self.scheduled_event_id is not None:
            url += "?event=" + str(self.scheduled_event_id)
        return url

    def set_scheduled_event(self, scheduled_event: Snowflake, /) -> Self:
        """Sets the scheduled event for this invite.

        .. versionadded:: 2.0

        Parameters
        ----------
        scheduled_event: :class:`~discord.abc.Snowflake`
            The ID of the scheduled event.

        Returns
        --------
        :class:`Invite`
            The invite with the new scheduled event.
        """
        self.scheduled_event_id = scheduled_event.id
        try:
            self.scheduled_event = self.guild.get_scheduled_event(scheduled_event.id)  # type: ignore # handled below
        except AttributeError:
            self.scheduled_event = None

        return self

    async def use(self) -> Invite:
        """|coro|

        Uses the invite.
        Either joins a guild, joins a group DM, or adds a friend.

        There is an alias for this called :func:`accept`.

        .. versionadded:: 1.9

        Raises
        ------
        HTTPException
            Using the invite failed.

        Returns
        -------
        :class:`Invite`
            The accepted invite.
        """
        state = self._state
        type = self.type
        if message := self._message:
            kwargs = {"message": message}
        else:
            kwargs = {
                "guild_id": getattr(self.guild, "id", MISSING),
                "channel_id": getattr(self.channel, "id", MISSING),
                "channel_type": getattr(self.channel, "type", MISSING),
            }
        data = await state.http.accept_invite(self.code, type, **kwargs)
        return Invite.from_incomplete(state=state, data=data, message=message)

    async def accept(self) -> Invite:
        """|coro|

        Uses the invite.
        Either joins a guild, joins a group DM, or adds a friend.

        This is an alias of :func:`use`.

        .. versionadded:: 1.9

        Raises
        ------
        HTTPException
            Using the invite failed.

        Returns
        -------
        :class:`Invite`
            The accepted invite.
        """
        return await self.use()

    async def delete(self, *, reason: Optional[str] = None) -> Invite:
        """|coro|

        Revokes the instant invite.

        In a guild context, you must have :attr:`~Permissions.manage_channels` to do this.

        .. versionchanged:: 2.0

            The function now returns the deleted invite.

        Parameters
        -----------
        reason: Optional[:class:`str`]
            The reason for deleting this invite. Shows up on the audit log.

            Only applicable to guild invites.

        Raises
        -------
        Forbidden
            You do not have permissions to revoke invites.
        NotFound
            The invite is invalid or expired.
        HTTPException
            Revoking the invite failed.

        Returns
        --------
        :class:`Invite`
            The deleted invite.
        """
        state = self._state
        data = await state.http.delete_invite(self.code, reason=reason)
        return Invite.from_incomplete(state=state, data=data)
